#coverage:ignore
""" Determine costs for THC decomposition in QC """
from typing import Tuple
import numpy as np
from numpy.lib.scimath import arccos, arcsin  # has analytc continuatn to cplx
from openfermion.resource_estimates.utils import QR, QI


def compute_cost(n: int,
                 lam: float,
                 dE: float,
                 chi: int,
                 beta: int,
                 M: int,
                 stps: int,
                 verbose: bool = False) -> Tuple[int, int, int]:
    """ Determine fault-tolerant costs using THC decomposition in quantum chem

    Args:
        n (int) - the number of spin-orbitals
        lam (float) - the lambda-value for the Hamiltonian
        dE (float) - allowable error in phase estimation
        chi (int) - equivalent to aleph in the document, the number of bits for
            the representation of the coefficients
        beta (int) - equivalent to beth in the document, the number of bits for
            the rotations
        M (int) - the dimension for the THC decomposition
        stps (int) - an approximate number of steps to choose the precision of
            single qubit rotations in preparation of the equal superpositn state

    Returns:
        step_cost (int) - Toffolis per step
        total_cost (int) - Total number of Toffolis
        ancilla_cost (int) - Total ancilla cost
    """
    # The number of bits used for each register.
    nM = np.ceil(np.log2(M + 1))

    # This is the number of distinct items of data we need to output.
    d = M * (M + 1) // 2 + n // 2

    # The number of bits used for the contiguous register.
    nc = np.ceil(np.log2(d))

    # The output size is 2*log[M] for the alt values, chi for the keep value,
    # and 2 for the two sign bits, as in Eq. (29).
    m = 2 * nM + 2 + chi
    oh = [0] * 20
    for p in range(20):
        # arccos arg may be > 1
        v = np.round(
            np.power(2, p + 1) / (2 * np.pi) *
            arccos(np.power(2, nM) / np.sqrt(d) / 2))
        oh[p] = stps * (1 / (np.sin(3 * arcsin(np.cos(v * 2 * np.pi / \
            np.power(2,p+1)) * \
            np.sqrt(d) / np.power(2,nM)))**2) - 1) + 4 * (p + 1)

    # Set it to be the number of bits that minimises the cost, usually 7.
    # Python is 0-index, so need to add the one back in vs mathematica nb
    br = np.argmin(oh) + 1

    # This is the costing for preparing the equal superposition over the
    # input registers from below Eq. (27).
    cp1 = 2 * (10 * nM + 2 * br - 9)

    # This is the cost of computing the contiguous register and inverting it.
    # This is with a sophisticated scheme adding together triplets of bits.
    # This is the cost of step 2 in the list on pages 15 and 16, with a factor
    # of 2 to account for its inverse.
    cp2 = 2 * (nM**2 + nM - 1)

    # This is the cost of the QROM for the state preparation in step 3 and its
    # inverse. Note: arg_min is first value, min is second value
    cp3 = QR(d, m)[1] + QI(d)[1]

    # The cost for the inequality test in step 4 and its inverse.
    cp4 = 2 * chi

    # The cost 2*nM for the controlled swap in step 5 and its inverse.
    cp5 = 4 * nM

    # Then there is a cost of nM + 1 for swapping the mu and nu registers in
    # step 6, where the + 1 is because we need to control on two registers.
    # There is the same cost for the inverse.
    cp6 = 2 * nM + 2

    # This is the total cost in Eq. (32).
    CPCP = cp1 + cp2 + cp3 + cp4 + cp5 + cp6

    # This is the cost of swapping based on the spin register. It is steps 1
    # and 7 from the list of steps on pages 16 and 17.
    cs1 = 2 * n

    # The QROM for the rotation angles the first time in step 2. and the second
    # part is the cost for inverting them with advanced QROM (step 6).
    cs2a = M + n / 2 - 2

    # The QROM for the rotation angles the second time. Here the cost M - 2 is
    # for generating the angles the second time (again step 2), and QI[M] is
    # for inverting the QROM (step 6).
    cs2b = M - 2

    # The cost of the rotations in steps 3 and 5, which must be done 4 times.
    cs3 = 4 * n * (beta - 2)

    # Cost for making the Z doubly controlled in step 4, and a controlled swap
    #  of the spin qubits. Note: is this really just "2"? Does this need a var?
    cs4 = 2

    # want the argument, so that is first element
    k1 = np.power(2, QI(M + n // 2)[0])

    # The cost for inverting the rotation angles the first time in step 6.
    cs6a = np.ceil(M / k1) + np.ceil(n / 2 / k1) + k1

    # The cost for inverting the rotation angles the second time in step 6.
    # value is in second output argument of QI
    cs6b = QI(M)[1]

    # The total select cost in Eq. (42).
    CS = cs1 + cs2a + cs2b + cs3 + cs4 + cs6a + cs6b

    # The cost of the reflections used in Eq. (43).
    costref = 2 * nM + chi + 4

    # The total cost in Eq. (43).
    cost = CPCP + CS + costref

    # The cost of the control register and temporary registers given for part 1.
    ac1 = lambda iters: 2 * np.ceil(np.log2(iters + 1)) - 1

    # Qubits for the system register in part 2.
    ac2 = n

    # The cost for the \[Mu] and \[Nu] registers in part 3.
    ac3 = 2 * nM

    # This is for the equal superposition state to perform the inequal
    # test with the keep register.
    ac4 = chi

    # Some minor qubit counts in parts 5 to 9.
    ac59 = 7

    # The phase gradient state in part 10.
    ac10 = beta

    # The contiguous register in part 11
    ac11 = nc

    # we want arg_min so first argument
    kt = np.power(2, QR(d, m))[0]

    # The qubits used for the QROM in part 12, of which only m are persistent.
    ac12 = m * kt + np.ceil(np.log2(d / kt))

    # This is the data needed for the rotations in part 13.
    ac13 = beta * n / 2

    # These are the ancillas needed for adding into the phase gradient state
    # in part 14.
    ac14 = beta - 2

    # These are the temporary ancillas in between erasing the first QROM
    # ancillas and inverting that QROM. The + m is for output of the first QROM.
    acc = ac13 + ac14 + m

    # Total number of iterations
    iters = np.ceil(np.pi * lam / (dE * 2))
    aca = ac1(iters) + ac2 + ac3 + ac4 + ac59 + ac10 + ac11

    if verbose:
        print("[*] Top of routine")
        print("  [+] nM = ", nM)
        print("  [+] d = ", d)
        print("  [+] nc = ", nc)
        print("  [+] m = ", m)
        print("  [+] oh = ", oh)
        print("  [+] br = ", br)
        print("  [+] cp1 = ", cp1)
        print("  [+] cp2 = ", cp2)
        print("    [*] QR[d,m] = ", QR(d, m))
        print("    [*] QI[d] = ", QI(d))
        print("  [+] cp3 = ", cp3)
        print("  [+] cp4 = ", cp4)
        print("  [+] cp5 = ", cp5)
        print("  [+] cp6 = ", cp6)
        print("  [+] CPCP = ", CPCP)
        print("  [+] cs1 = ", cs1)
        print("  [+] cs2a = ", cs2a)
        print("  [+] cs2b = ", cs2b)
        print("  [+] cs3 = ", cs3)
        print("  [+] k1 = ", k1)
        print("  [+] cs6a = ", cs6a)
        print("  [+] cs6b = ", cs6b)
        print("  [+] CS = ", CS)
        print("  [+] costref = ", costref)
        print("  [+] cost = ", cost)
        print("  [+] ac1 = ", ac1)
        print("  [+] ac2 = ", ac2)
        print("  [+] ac3 = ", ac3)
        print("  [+] ac4 = ", ac4)
        print("  [+] ac10 = ", ac10)
        print("  [+] ac11 = ", ac11)
        print("  [+] kt = ", kt)
        print("  [+] ac12 = ", ac12)
        print("  [+] ac13 = ", ac13)
        print("  [+] ac14 = ", ac14)
        print("  [+] acc = ", acc)
        print("  [+] iters = ", iters)
        print("  [+] aca = ", aca)

    # Sanity checks before returning as int
    # assert cost.is_integer()
    # assert iters.is_integer()
    # assert aca.is_integer()
    # assert ac12.is_integer()
    # assert acc.is_integer()

    step_cost = int(cost)
    total_cost = int(cost * iters)
    ancilla_cost = int(np.max([aca + ac12, aca + acc]))

    # step-cost, Toffoli count, logical qubits
    return step_cost, total_cost, ancilla_cost



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import numpy.typing as npt
"""Module for performing K-Means CVT algorithm to find interpolating points.

Provides centroidal Veronoi tesselation of the grid of real space points
weighted by the electron density using a K-Means classification.
"""


class KMeansCVT(object):

    def __init__(
            self,
            grid: npt.NDArray,
            max_iteration: int = 100,
            threshold: float = 1e-6,
    ):
        """Initialize k-means solver to find interpolating points for ISDF.

        Args:
        grid: Real space grid of dimension [Ng,Ndim], where Ng is the number
            of (dense) real space grid points and Ndim is number of spatial
            dimensions.
        max_iteration: Maximum number of iterations to perform when
            classifying grid points. Default 100.
        threshold: Threshold for exiting classification. Default 1e-6.

        Returns:
        """
        self.grid = grid
        self.max_iteration = max_iteration
        self.threshold = threshold

    @staticmethod
    def classify_grid_points(grid_points: npt.NDArray,
                             centroids: npt.NDArray) -> npt.NDArray:
        r"""Assign grid points to centroids.

        Find centroid closest to each given grid point.

        Note we don't use instance variable self.grid as we can abuse this
        function and use it to map grid to centroid and centroid to grid point.

        Args:
          grid_points: grid points to assign.
          centroids: Centroids to which grid points should be assigned,
            array of length num_interp_points.

        Returns:
          1D np.array assigning grid point to centroids
        """
        # Build N_g x N_mu matrix of distances.
        num_grid_points = grid_points.shape[0]
        num_interp_points = centroids.shape[0]
        distances = np.zeros((num_grid_points, num_interp_points))
        # For loop is faster than broadcasting by 2x.
        for ig in range(num_grid_points):
            distances[ig] = np.linalg.norm(grid_points[ig] - centroids, axis=1)
        # Find shortest distance for each grid point.
        classification = np.argmin(distances, axis=1)
        return classification

    def compute_new_centroids(self, weighting, grid_mapping,
                              current_centroids) -> npt.NDArray:
        r"""
        Centroids are defined via:

        .. math::

            c(C_\mu) = \frac{\sum_{j in C(\mu)} r_j \rho(r_j)}{\sum_{j in
                C(\mu)} \rho(r_j)},

        where :math:`\rho(r_j)` is the weighting factor.

        Args:
            weighting: Weighting function.
            grid_mapping: maps grid points to centroids
            current_centroids: centroids for current iteration.

        Returns:
            new_centroids: updated centroids
        """
        num_interp_points = current_centroids.shape[0]
        new_centroids = np.zeros_like(current_centroids)
        for interp_indx in range(num_interp_points):
            # get grid points belonging to this centroid
            grid_indx = np.where(grid_mapping == interp_indx)[0]
            grid_points = self.grid[grid_indx]
            weight = weighting[grid_indx]
            numerator = np.einsum("Jx,J->x", grid_points, weight)
            denominator = np.einsum("J->", weight)
            if denominator < 1e-12:
                print("Warning very small denominator, something seems wrong!")
                print("{interp_indx}")
            new_centroids[interp_indx] = numerator / denominator
        return new_centroids

    def map_centroids_to_grid(self, centroids):
        grid_mapping = self.classify_grid_points(centroids, self.grid)
        return grid_mapping

    def find_interpolating_points(
            self,
            num_interp_points: int,
            weighting_factor: npt.NDArray,
            centroids=None,
            verbose=True,
    ) -> npt.NDArray:
        """Find interpolating points using KMeans-CVT algorithm.

        Args:
            num_interp_points: number of points to select.
            weighting_factor: weighting function for K-Means procedure.
            centroids: initial guess at centroids, if None centroids are
                selected randomly from the grid points.
            verbose: Controls if information is printed about convergence.
                Default value = True.

        Returns:
            interp_pts: index associated with interpolating points.
        """
        num_grid_points = self.grid.shape[0]
        if centroids is None:
            # Randomly select grid points as centroids.
            centroids_indx = np.random.choice(num_grid_points,
                                              num_interp_points,
                                              replace=False)
            centroids = self.grid[centroids_indx].copy()
        else:
            assert len(centroids) == num_interp_points
        # define here to avoid linter errors about possibly undefined.
        new_centroids = np.zeros_like(centroids)
        delta_grid = 1.0
        if verbose:
            print("{:<10s}  {:>13s}".format("iteration", "Error"))
        for iteration in range(self.max_iteration):
            grid_mapping = self.classify_grid_points(self.grid, centroids)
            # Global reduce
            new_centroids[:] = self.compute_new_centroids(
                weighting_factor, grid_mapping, centroids)
            delta_grid = np.linalg.norm(new_centroids - centroids)
            if verbose and iteration % 10 == 0:
                print(f"{iteration:<9d}  {delta_grid:13.8e}")
            if delta_grid < self.threshold:
                if verbose:
                    print("KMeansCVT successfully completed.")
                    print(f"Final error {delta_grid:13.8e}.")
                return self.map_centroids_to_grid(new_centroids)
            centroids[:] = new_centroids[:]
        print("Warning K-Means not converged.")
        print(f"Final error {delta_grid:13.8e}.")
        return self.map_centroids_to_grid(new_centroids)



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import itertools

import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from openfermion.resource_estimates.pbc.thc.factorizations.gvec_map_logic import (  # pylint: disable=line-too-long
        build_conjugate_map, build_G_vectors, build_gpq_mapping,
        build_transfer_map, get_delta_kp_kq_q, get_miller_indices)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_get_miller_indices():
    kmesh = [3, 1, 1]
    int_scaled_kpts = get_miller_indices(kmesh)
    assert np.allclose(int_scaled_kpts[:, 0], np.arange(3))
    assert np.allclose(int_scaled_kpts[:, 1], 0)
    assert np.allclose(int_scaled_kpts[:, 2], 0)

    kmesh = [3, 2, 1]
    int_scaled_kpts = get_miller_indices(kmesh)
    assert np.allclose(int_scaled_kpts[:, 0], [0, 0, 1, 1, 2, 2])
    assert np.allclose(int_scaled_kpts[:, 1], [0, 1, 0, 1, 0, 1])


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_get_delta_k1_k2_Q():
    kmesh = [3, 2, 1]
    nkpts = np.prod(kmesh)
    scaled_kpts = get_miller_indices(kmesh)
    delta_k1_k2_Q_int = get_delta_kp_kq_q(scaled_kpts)
    assert delta_k1_k2_Q_int.shape == (nkpts, nkpts, nkpts, 3)
    for kpidx, kqidx, qidx in itertools.product(range(nkpts), repeat=3):
        assert np.allclose(
            scaled_kpts[kpidx] - scaled_kpts[kqidx] - scaled_kpts[qidx],
            delta_k1_k2_Q_int[kpidx, kqidx, qidx],
        )


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_transfer_map():
    kmesh = [3, 1, 1]
    nkpts = np.prod(kmesh)
    scaled_kpts = get_miller_indices(kmesh)
    transfer_map = build_transfer_map(kmesh, scaled_kpts=scaled_kpts)
    for kpidx, kqidx in itertools.product(range(nkpts), repeat=2):
        q_scaled_kpt = scaled_kpts[kpidx] - scaled_kpts[kqidx]
        q_scaled_kpt_aliased = np.array([
            q_scaled_kpt[0] % kmesh[0],
            q_scaled_kpt[1] % kmesh[1],
            q_scaled_kpt[2] % kmesh[2],
        ])
        qidx = np.where((scaled_kpts[:, 0] == q_scaled_kpt_aliased[0]) &
                        (scaled_kpts[:, 1] == q_scaled_kpt_aliased[1]) &
                        (scaled_kpts[:, 2] == q_scaled_kpt_aliased[2]))[0]
        assert transfer_map[qidx, kpidx] == kqidx

    true_transfer_map = np.array([[0, 1, 2], [2, 0, 1], [1, 2, 0]])
    assert np.allclose(transfer_map, true_transfer_map)
    kmesh = [4, 1, 1]
    nkpts = np.prod(kmesh)
    scaled_kpts = get_miller_indices(kmesh)
    transfer_map = build_transfer_map(kmesh, scaled_kpts=scaled_kpts)
    for kpidx, kqidx in itertools.product(range(nkpts), repeat=2):
        q_scaled_kpt = scaled_kpts[kpidx] - scaled_kpts[kqidx]
        q_scaled_kpt_aliased = np.array([
            q_scaled_kpt[0] % kmesh[0],
            q_scaled_kpt[1] % kmesh[1],
            q_scaled_kpt[2] % kmesh[2],
        ])
        qidx = np.where((scaled_kpts[:, 0] == q_scaled_kpt_aliased[0]) &
                        (scaled_kpts[:, 1] == q_scaled_kpt_aliased[1]) &
                        (scaled_kpts[:, 2] == q_scaled_kpt_aliased[2]))[0]
        assert transfer_map[qidx, kpidx] == kqidx

    true_transfer_map = np.array([[0, 1, 2, 3], [3, 0, 1, 2], [2, 3, 0, 1],
                                  [1, 2, 3, 0]])
    assert np.allclose(transfer_map, true_transfer_map)

    kmesh = [3, 2, 1]
    nkpts = np.prod(kmesh)
    scaled_kpts = get_miller_indices(kmesh)
    transfer_map = build_transfer_map(kmesh, scaled_kpts=scaled_kpts)
    for kpidx, kqidx in itertools.product(range(nkpts), repeat=2):
        q_scaled_kpt = scaled_kpts[kpidx] - scaled_kpts[kqidx]
        q_scaled_kpt_aliased = np.array([
            q_scaled_kpt[0] % kmesh[0],
            q_scaled_kpt[1] % kmesh[1],
            q_scaled_kpt[2] % kmesh[2],
        ])
        qidx = np.where((scaled_kpts[:, 0] == q_scaled_kpt_aliased[0]) &
                        (scaled_kpts[:, 1] == q_scaled_kpt_aliased[1]) &
                        (scaled_kpts[:, 2] == q_scaled_kpt_aliased[2]))[0]
        assert transfer_map[qidx, kpidx] == kqidx

    true_transfer_map = np.array([
        [0, 1, 2, 3, 4, 5],
        [1, 0, 3, 2, 5, 4],
        [4, 5, 0, 1, 2, 3],
        [5, 4, 1, 0, 3, 2],
        [2, 3, 4, 5, 0, 1],
        [3, 2, 5, 4, 1, 0],
    ])
    assert np.allclose(transfer_map, true_transfer_map)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_build_Gvectors():
    kmesh = [3, 2, 1]
    g_dict = build_G_vectors(kmesh)
    indx = 0
    for n1, n2, n3 in itertools.product([0, -1], repeat=3):
        assert np.isclose(g_dict[(n1 * kmesh[0], n2 * kmesh[1], n3 * kmesh[2])],
                          indx)
        indx += 1


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_gpq_mapping():
    kmesh = [3, 2, 1]
    nkpts = np.prod(kmesh)
    scaled_kpts = get_miller_indices(kmesh)
    transfer_map = build_transfer_map(kmesh, scaled_kpts=scaled_kpts)
    gpq_map = build_gpq_mapping(kmesh, scaled_kpts)
    for iq in range(nkpts):
        for ikp in range(nkpts):
            ikq = transfer_map[iq, ikp]
            q_plus_g = scaled_kpts[ikp] - scaled_kpts[ikq]
            g_val = gpq_map[ikp, ikq]
            q_val = q_plus_g - np.array(g_val)
            assert np.allclose(q_val, scaled_kpts[iq])

            assert all(g_val <= 0)
            assert g_val[0] in [0, -kmesh[0]]
            assert g_val[1] in [0, -kmesh[1]]
            assert g_val[2] in [0, -kmesh[2]]


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_build_conjugate_map():
    kmesh = [4, 3, 3]
    nkpts = np.prod(kmesh)
    scaled_kpts = get_miller_indices(kmesh)
    kconj_map = build_conjugate_map(kmesh, scaled_kpts=scaled_kpts)
    for kidx in range(nkpts):
        neg_kscaled_idx = kconj_map[kidx]
        gval = scaled_kpts[neg_kscaled_idx] + scaled_kpts[kidx]
        # -Q + Q = 0 + G
        assert gval[0] in [0, kmesh[0]]
        assert gval[1] in [0, kmesh[1]]
        assert gval[2] in [0, kmesh[2]]


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_compliment_g():
    # setup
    kmesh = [4, 1, 1]
    nkpts = np.prod(kmesh)
    scaled_kpts = get_miller_indices(kmesh)
    g_dict = build_G_vectors(kmesh)
    gpq_map = build_gpq_mapping(kmesh, scaled_kpts)
    transfer_map = build_transfer_map(kmesh, scaled_kpts=scaled_kpts)

    # confirm that our transfer map function is working
    for qidx, kpidx, ksidx in itertools.product(range(nkpts), repeat=3):
        kqidx = transfer_map[qidx, kpidx]
        kridx = transfer_map[qidx, ksidx]

        conserved_k = (scaled_kpts[kpidx] - scaled_kpts[kqidx] +
                       scaled_kpts[kridx] - scaled_kpts[ksidx])
        assert conserved_k[0] % kmesh[0] == 0
        assert conserved_k[1] % kmesh[1] == 0
        assert conserved_k[2] % kmesh[2] == 0

    # generate set of k for every (q and g)
    q_idx_g_idx = {}
    for qidx in range(nkpts):
        for gvals, gidx in g_dict.items():
            q_idx_g_idx[(qidx, gidx)] = []
            # print("Traget gval ", gvals)
            kidx_satisfying_q_g = []
            for kidx in range(nkpts):
                kmq_idx = transfer_map[qidx, kidx]
                test_gval = gpq_map[kidx, kmq_idx]
                # print(test_gval, test_gval == gvals)
                if all(test_gval == gvals):
                    kidx_satisfying_q_g.append(kidx)
            q_idx_g_idx[(qidx, gidx)] = kidx_satisfying_q_g

    # count how many total k-values and how many per q values
    total_k = 0
    for qidx in range(nkpts):
        sub_val = 0
        for g1idx, g2idx in itertools.product(range(8), repeat=2):
            total_k += len(q_idx_g_idx[(qidx, g1idx)]) * len(
                q_idx_g_idx[(qidx, g2idx)])
            sub_val += len(q_idx_g_idx[(qidx, g1idx)]) * len(
                q_idx_g_idx[(qidx, g2idx)])
        assert np.isclose(sub_val, nkpts**2)

    assert np.isclose(total_k, nkpts**3)



#coverage:ignore
# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Union
import numpy as np

Step = int
Schedule = Callable[[Step], float]


def adagrad(step_size, momentum=0.9):
    """Construct optimizer triple for Adagrad.
  Adaptive Subgradient Methods for Online Learning and Stochastic Optimization:
  http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf
  Args:
    step_size: positive scalar, or a callable representing a step size schedule
      that maps the iteration index to positive scalar.
    momentum: optional, a positive scalar value for momentum
  Returns:
    An (init_fun, update_fun, get_params) triple.
  """
    step_size = make_schedule(step_size)

    def init(x0):
        g_sq = np.zeros_like(x0)
        m = np.zeros_like(x0)
        return x0, g_sq, m

    def update(i, g, state):
        x, g_sq, m = state
        g_sq += np.square(g)
        g_sq_inv_sqrt = np.where(g_sq > 0, 1. / np.sqrt(g_sq), 0.0)
        m = (1. - momentum) * (g * g_sq_inv_sqrt) + momentum * m
        x = x - step_size(i) * m
        return x, g_sq, m

    def get_params(state):
        x, _, _ = state
        return x

    return init, update, get_params


### learning rate schedules


def constant(step_size) -> Schedule:

    def schedule(i):
        return step_size

    return schedule


def exponential_decay(step_size, decay_steps, decay_rate):

    def schedule(i):
        return step_size * decay_rate**(i / decay_steps)

    return schedule


def inverse_time_decay(step_size, decay_steps, decay_rate, staircase=False):
    if staircase:

        def schedule(i):
            return step_size / (1 + decay_rate * np.floor(i / decay_steps))
    else:

        def schedule(i):
            return step_size / (1 + decay_rate * i / decay_steps)

    return schedule


def polynomial_decay(step_size, decay_steps, final_step_size, power=1.0):

    def schedule(step_num):
        step_num = np.minimum(step_num, decay_steps)
        step_mult = (1 - step_num / decay_steps)**power
        return step_mult * (step_size - final_step_size) + final_step_size

    return schedule


def piecewise_constant(boundaries: Any, values: Any):
    boundaries = np.array(boundaries)
    values = np.array(values)
    if not boundaries.ndim == values.ndim == 1:
        raise ValueError("boundaries and values must be sequences")
    if not boundaries.shape[0] == values.shape[0] - 1:
        raise ValueError(
            "boundaries length must be one shorter than values length")

    def schedule(i):
        return values[np.sum(i > boundaries)]

    return schedule


def make_schedule(scalar_or_schedule: Union[float, Schedule]) -> Schedule:
    if callable(scalar_or_schedule):
        return scalar_or_schedule
    elif np.ndim(scalar_or_schedule) == 0:
        return constant(scalar_or_schedule)
    else:
        raise TypeError(type(scalar_or_schedule))



#coverage:ignore
"""Test cases for costing_thc.py
"""
import unittest

import pytest

from openfermion.resource_estimates import thc


class THCCostTest(unittest.TestCase):

    def test_reiher_thc(self):
        """ Reproduce Reiher et al orbital THC FT costs from paper """
        DE = 0.001
        CHI = 10

        # Reiher et al orbitals
        N = 108
        LAM = 306.3
        BETA = 16
        THC_DIM = 350

        # Here we're using an initial calculation with a very rough estimate of
        # the number of steps to give a more accurate number of steps. Then we
        # input that into the function again.
        output = thc.compute_cost(N, LAM, DE, CHI, BETA, THC_DIM, stps=20000)
        stps1 = output[0]
        output = thc.compute_cost(N, LAM, DE, CHI, BETA, THC_DIM, stps1)
        assert output == (10912, 5250145120, 2142)

    def test_li_thc(self):
        """ Reproduce Li et al orbital THC FT costs from paper """
        DE = 0.001
        CHI = 10

        # Li et al orbitals
        N = 152
        LAM = 1201.5
        BETA = 20
        THC_DIM = 450

        # Here we're using an initial calculation with a very rough estimate of
        # the number of steps to give a more accurate number of steps. Then we
        # input that into the function again.
        output = thc.compute_cost(N, LAM, DE, CHI, BETA, THC_DIM, stps=20000)
        stps2 = output[0]
        output = thc.compute_cost(N, LAM, DE, CHI, BETA, THC_DIM, stps2)
        assert output == (16923, 31938980976, 2196)



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Dataclasses for resource estimation."""
from dataclasses import dataclass, asdict, field
import pandas as pd

from openfermion.resource_estimates.pbc.hamiltonian import (
    HamiltonianProperties,)


@dataclass(frozen=True)
class ResourceEstimates:
    """Dataclass to hold return values from compute_cost functions.

    Attributes:
        toffolis_per_step: Toffolis per step
        total_toffolis: Total number of Toffolis
        logical_qubits: Total ancilla cost
    """

    toffolis_per_step: int
    total_toffolis: int
    logical_qubits: int

    dict = asdict


@dataclass
class PBCResources:
    """Helper class to hold resource estimates given a range of cutoffs.

    Attributes:
        system_name: Descriptive name for calculation.
        num_spin_orbitals: Number of spin orbitals.
        num_kpts: Number of k-points.
        dE: Epsilon for phase estimation.
        chi: A number of bits controlling the precision of the factorization.
            What this means is dependent on the factorization.
        exact_energy: Exact energy (no cutoff).
        energy_method: Which method to use to copute energy (MP2 or CCSD).
        cutoff: list of cuttofs.
        ham_props: list of lambda values for each cutoff.
        resources: list of resource estimates for each cutoff.
    """

    system_name: str
    num_spin_orbitals: int
    num_kpts: int
    dE: float
    chi: int
    exact_energy: float
    energy_method: str = "MP2"
    cutoff: list = field(default_factory=list)
    approx_energy: list = field(default_factory=list)
    ham_props: list = field(default_factory=list)
    resources: list = field(default_factory=list)

    dict = asdict

    def to_dataframe(self) -> pd.DataFrame:
        """Convert PBCResources instance to pandas DataFrame."""
        df = pd.DataFrame(self.dict())
        lambdas = pd.json_normalize(df.pop("ham_props"))
        resources = pd.json_normalize(df.pop("resources"))
        df = df.join(pd.DataFrame(lambdas))
        df = df.join(pd.DataFrame(resources))
        return df

    def add_resources(
            self,
            ham_properties: HamiltonianProperties,
            resource_estimates: ResourceEstimates,
            cutoff: float,
            approx_energy: float,
    ) -> None:
        """Add resource estimates to container for given cutoff value.

        Args:
            ham_properties: lambda values
            resource_estimates: resource estimates for this value of cutoff
            cutoff: Current factorization/representation cutoff value.
            approx_energy: Current approximate energy estimate given cutoff.
        """
        self.ham_props.append(ham_properties)
        self.resources.append(resource_estimates)
        self.cutoff.append(cutoff)
        self.approx_energy.append(approx_energy)



#coverage:ignore
""" THC rank reduction of ERIs """
import time

import h5py
import numpy as np

from openfermion.resource_estimates.thc.utils import lbfgsb_opt_thc_l2reg


def thc_via_cp3(eri_full,
                nthc,
                thc_save_file=None,
                first_factor_thresh=1.0E-14,
                conv_eps=1.0E-4,
                perform_bfgs_opt=True,
                bfgs_maxiter=5000,
                random_start_thc=True,
                verify=False):
    """
    THC-CP3 performs an SVD decomposition of the eri matrix followed by a CP
    decomposition via pybtas. The CP decomposition is assumes the tensor is
    symmetric in the first two indices corresponding to a reshaped
    (and rescaled by the singular value) singular vector.

    Args:
        eri_full - (N x N x N x N) eri tensor in Mulliken (chemists) ordering
        nthc (int) - number of THC factors to use
        thc_save_file (str) - if not None, save output to filename (as HDF5)
        first_factor_thresh - SVD threshold on initial factorization of ERI
        conv_eps (float) - convergence threshold on CP3 ALS
        perform_bfgs_opt - Perform extra gradient opt. on top of CP3 decomp
        bfgs_maxiter - Maximum bfgs steps to take. Default 1500.
        random_start_thc - Perform random start for CP3.
                           If false perform HOSVD start.
        verify - check eri properties. Default is False

    returns:
        eri_thc - (N x N x N x N) reconstructed ERIs from THC factorization
        thc_leaf - THC leaf tensor
        thc_central - THC central tensor
        info (dict) - arguments set during the THC factorization
    """
    # fail fast if we don't have the tools to use this routine
    try:
        import pybtas
    except ImportError:
        raise ImportError(
            "pybtas could not be imported. Is it installed and in PYTHONPATH?")

    info = locals()
    info.pop('eri_full', None)  # data too big for info dict
    info.pop('pybtas', None)  # not needed for info dict

    norb = eri_full.shape[0]
    if verify:
        assert np.allclose(eri_full,
                           eri_full.transpose(1, 0, 2, 3))  # (ij|kl) == (ji|kl)
        assert np.allclose(eri_full,
                           eri_full.transpose(0, 1, 3, 2))  # (ij|kl) == (ij|lk)
        assert np.allclose(eri_full,
                           eri_full.transpose(1, 0, 3, 2))  # (ij|kl) == (ji|lk)
        assert np.allclose(eri_full,
                           eri_full.transpose(2, 3, 0, 1))  # (ij|kl) == (kl|ij)

    eri_mat = eri_full.transpose(0, 1, 3, 2).reshape((norb**2, norb**2))
    if verify:
        assert np.allclose(eri_mat, eri_mat.T)

    u, sigma, vh = np.linalg.svd(eri_mat)

    if verify:
        assert np.allclose(u @ np.diag(sigma) @ vh, eri_mat)

    non_zero_sv = np.where(sigma >= first_factor_thresh)[0]
    u_chol = u[:, non_zero_sv]
    diag_sigma = np.diag(sigma[non_zero_sv])
    u_chol = u_chol @ np.diag(np.sqrt(sigma[non_zero_sv]))

    if verify:
        test_eri_mat_mulliken = u[:,
                                  non_zero_sv] @ diag_sigma @ vh[non_zero_sv, :]
        assert np.allclose(test_eri_mat_mulliken, eri_mat)

    start_time = time.time()  # timing results if requested by user
    beta, gamma, scale = pybtas.cp3_from_cholesky(u_chol.copy(),
                                                  nthc,
                                                  random_start=random_start_thc,
                                                  conv_eps=conv_eps)
    cp3_calc_time = time.time() - start_time

    if verify:
        u_alpha = np.zeros((norb, norb, len(non_zero_sv)))
        for ii in range(len(non_zero_sv)):
            u_alpha[:, :, ii] = np.sqrt(sigma[ii]) * u[:, ii].reshape(
                (norb, norb))
            assert np.allclose(
                u_alpha[:, :, ii],
                u_alpha[:, :, ii].T)  # consequence of working with Mulliken rep

        u_alpha_test = np.einsum("ar,br,xr,r->abx", beta, beta, gamma,
                                 scale.ravel())
        print("\tu_alpha l2-norm ", np.linalg.norm(u_alpha_test - u_alpha))

    thc_leaf = beta.T
    thc_gamma = np.einsum('xr,r->xr', gamma, scale.ravel())
    thc_central = thc_gamma.T @ thc_gamma

    if verify:
        eri_thc = np.einsum("Pp,Pr,Qq,Qs,PQ->prqs",
                            thc_leaf,
                            thc_leaf,
                            thc_leaf,
                            thc_leaf,
                            thc_central,
                            optimize=True)
        print("\tERI L2 CP3-THC ", np.linalg.norm(eri_thc - eri_full))
        print("\tCP3 timing: ", cp3_calc_time)

    if perform_bfgs_opt:
        x = np.hstack((thc_leaf.ravel(), thc_central.ravel()))
        #lbfgs_start_time = time.time()
        x = lbfgsb_opt_thc_l2reg(eri_full,
                                 nthc,
                                 initial_guess=x,
                                 maxiter=bfgs_maxiter)
        #lbfgs_calc_time = time.time() - lbfgs_start_time
        thc_leaf = x[:norb * nthc].reshape(nthc,
                                           norb)  # leaf tensor  nthc x norb
        thc_central = x[norb * nthc:norb * nthc + nthc * nthc].reshape(
            nthc, nthc)  # central tensor

    #total_calc_time = time.time() - start_time

    eri_thc = np.einsum("Pp,Pr,Qq,Qs,PQ->prqs",
                        thc_leaf,
                        thc_leaf,
                        thc_leaf,
                        thc_leaf,
                        thc_central,
                        optimize=True)

    if thc_save_file is not None:
        with h5py.File(thc_save_file + '.h5', 'w') as fid:
            fid.create_dataset('thc_leaf', data=thc_leaf)
            fid.create_dataset('thc_central', data=thc_central)
            fid.create_dataset('info', data=str(info))

    return eri_thc, thc_leaf, thc_central, info



#coverage:ignore
""" Pretty-print a table comparing number of THC vectors vs accy and cost """
import numpy as np

from openfermion.resource_estimates import (HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                                            thc)

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from pyscf import scf

    from openfermion.resource_estimates.molecule import (cas_to_pyscf,
                                                         factorized_ccsd_t,
                                                         pyscf_to_cas)


def generate_costing_table(pyscf_mf,
                           name='molecule',
                           nthc_range=None,
                           dE=0.001,
                           chi=10,
                           beta=20,
                           save_thc=False,
                           use_kernel=True,
                           no_triples=False,
                           **kwargs):
    """ Print a table to file for testing how various THC thresholds impact
        cost, accuracy, etc.

    Args:
        pyscf_mf - PySCF mean field object
        name (str) - file will be saved to 'thc_factorization_<name>.txt'
        nthc_range (list of ints) - list of number of THC vectors to retain
        dE (float) - max allowable phase error (default: 0.001)
        chi (int) - number of bits for repr of coefficients (default: 10)
        beta (int) - number of bits for rotations (default: 20)
        save_thc (bool) - if True, save the THC factors (leaf and central only)
        use_kernel (bool) - re-do SCF prior to estimating CCSD(T) error?
            Will use canonical orbitals and full ERIs for the one-body
            contributions, using rank-reduced ERIs for two-body
        no_triples (bool) - if True, skip the (T) correction, doing only CCSD
        kwargs: additional keyword arguments to pass to thc.factorize()

    Returns:
       None
    """

    if nthc_range is None:
        nthc_range = [250, 300, 350]

    DE = dE  # max allowable phase error
    CHI = chi  # number of bits for representation of coefficients
    BETA = beta  # number of bits for the rotations

    if isinstance(pyscf_mf, scf.rohf.ROHF):
        num_alpha, num_beta = pyscf_mf.nelec
        assert num_alpha + num_beta == pyscf_mf.mol.nelectron
    else:
        assert pyscf_mf.mol.nelectron % 2 == 0
        num_alpha = pyscf_mf.mol.nelectron // 2
        num_beta = num_alpha

    num_orb = len(pyscf_mf.mo_coeff)
    num_spinorb = num_orb * 2

    cas_info = "CAS((%sa, %sb), %so)" % (num_alpha, num_beta, num_orb)

    try:
        assert num_orb**4 == len(pyscf_mf._eri.flatten())
    except AssertionError:
        # ERIs are not in correct form in pyscf_mf._eri, so this is a quick prep
        _, pyscf_mf = cas_to_pyscf(*pyscf_to_cas(pyscf_mf))

    # Reference calculation (eri_rr= None is full rank / exact ERIs)
    escf, ecor, etot = factorized_ccsd_t(pyscf_mf,
                                         eri_rr=None,
                                         use_kernel=use_kernel,
                                         no_triples=no_triples)

    exact_etot = etot

    filename = 'thc_factorization_' + name + '.txt'

    with open(filename, 'w') as f:
        print("\n THC factorization data for '" + name + "'.", file=f)
        print("    [*] using " + cas_info, file=f)
        print("        [+]                      E(SCF): %18.8f" % escf, file=f)
        if no_triples:
            print("        [+]    Active space CCSD E(cor): %18.8f" % ecor,
                  file=f)
            print("        [+]    Active space CCSD E(tot): %18.8f" % etot,
                  file=f)
        else:
            print("        [+] Active space CCSD(T) E(cor): %18.8f" % ecor,
                  file=f)
            print("        [+] Active space CCSD(T) E(tot): %18.8f" % etot,
                  file=f)
        print("{}".format('=' * 111), file=f)
        if no_triples:
            print("{:^12} {:^18} {:^24} {:^12} {:^20} {:^20}".format(
                'M', '||ERI - THC||', 'CCSD error (mEh)', 'lambda',
                'Toffoli count', 'logical qubits'),
                  file=f)
        else:
            print("{:^12} {:^18} {:^24} {:^12} {:^20} {:^20}".format(
                'M', '||ERI - THC||', 'CCSD(T) error (mEh)', 'lambda',
                'Toffoli count', 'logical qubits'),
                  file=f)
        print("{}".format('-' * 111), file=f)
    for nthc in nthc_range:
        # First, up: lambda and CCSD(T)
        if save_thc:
            fname = name + '_nTHC_' + str(nthc).zfill(
                5)  # will save as HDF5 and add .h5 extension
        else:
            fname = None
        eri_rr, thc_leaf, thc_central, info = thc.factorize(pyscf_mf._eri,
                                                            nthc,
                                                            thc_save_file=fname,
                                                            **kwargs)
        lam = thc.compute_lambda(pyscf_mf, thc_leaf, thc_central)[0]
        escf, ecor, etot = factorized_ccsd_t(pyscf_mf,
                                             eri_rr,
                                             use_kernel=use_kernel,
                                             no_triples=no_triples)
        error = (etot - exact_etot) * 1E3  # to mEh
        l2_norm_error_eri = np.linalg.norm(
            eri_rr - pyscf_mf._eri)  # ERI reconstruction error

        # now do costing
        stps1 = thc.compute_cost(num_spinorb,
                                 lam,
                                 DE,
                                 chi=CHI,
                                 beta=BETA,
                                 M=nthc,
                                 stps=20000)[0]
        _, thc_total_cost, thc_logical_qubits = thc.compute_cost(num_spinorb,
                                                                 lam,
                                                                 DE,
                                                                 chi=CHI,
                                                                 beta=BETA,
                                                                 M=nthc,
                                                                 stps=stps1)

        with open(filename, 'a') as f:
            print(
                "{:^12} {:^18.4e} {:^24.2f} {:^12.1f} {:^20.1e} {:^20}".format(
                    nthc, l2_norm_error_eri, error, lam, thc_total_cost,
                    thc_logical_qubits),
                file=f)
    with open(filename, 'a') as f:
        print("{}".format('=' * 111), file=f)

    with open(filename, 'a') as f:
        print("THC factorization settings at exit:", file=f)
        for key, value in info.items():
            print("\t", key, value, file=f)



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt
from openfermion.resource_estimates.pbc.hamiltonian import (
    HamiltonianProperties,)

from openfermion.resource_estimates.pbc.sf.sf_integrals import (
    SingleFactorization,)


@dataclass
class SFHamiltonianProperties(HamiltonianProperties):
    """Store for return values of compute_lambda function

    Extension of HamiltonianProperties dataclass to also hold the number of
    retained cholesky vectors (num_aux).
    """

    num_aux: int


def compute_lambda(hcore: npt.NDArray,
                   sf_obj: SingleFactorization) -> SFHamiltonianProperties:
    """Lambda for single-factorized Hamiltonian.

    Compute one-body and two-body lambda for qubitization of
    single-factorized Hamiltonian.

    one-body term h_pq(k) = hcore_{pq}(k)
                            - 0.5 * sum_{Q}sum_{r}(pkrQ|rQqk)
                            + sum_{Q}sum_{r}(pkqk|rQrQ)
    The first term is the kinetic energy + pseudopotential
    (or electron-nuclear), second term is from rearranging two-body operator
    into chemist charge-charge type notation, and the third is from the one body
    term obtained when squaring the two-body A and B operators.

    two-body term V = 0.5 sum_{Q}sum_{n}(A_{n}(Q)^2 +_ B_{n}(Q)^2)
    or V = 0.5 sum_{Q}sum_{n'}W_{n}(Q)^{2} where n' is twice the range of n.
    lambda is 0.5sum_{Q}sum_{n'}(sum_{p,q}^{N_{k}N/2}|Re[W_{p,q}(Q)^{n}]| +
    |Im[W_{pq}(Q)^{n}]|)^{2}

    Args:
        hcore: List len(kpts) long of nmo x nmo complex hermitian arrays
        sf_obj: SingleFactorization integral helper object.

    Returns:
        ham_props: A HamiltonianProperties instance containing Lambda values for
            SF hamiltonian.
    """
    kpts = sf_obj.kmf.kpts
    nkpts = len(kpts)
    one_body_mat = np.empty((len(kpts)), dtype=object)
    lambda_one_body = 0.0

    old_naux = sf_obj.naux  # need to reset naux for one-body computation
    sf_obj.naux = sf_obj.chol[0, 0].shape[0]

    for kidx in range(len(kpts)):
        # matrices for - 0.5 * sum_{Q}sum_{r}(pkrQ|rQqk)
        # and  + 0.5 sum_{Q}sum_{r}(pkqk|rQrQ)
        h1_pos = np.zeros_like(hcore[kidx])
        h1_neg = np.zeros_like(hcore[kidx])
        for qidx in range(len(kpts)):
            # - 0.5 * sum_{Q}sum_{r}(pkrQ|rQqk)
            eri_kqqk_pqrs = sf_obj.get_eri([kidx, qidx, qidx, kidx])
            h1_neg -= (np.einsum("prrq->pq", eri_kqqk_pqrs, optimize=True) /
                       nkpts)
            # + sum_{Q}sum_{r}(pkqk|rQrQ)
            eri_kkqq_pqrs = sf_obj.get_eri([kidx, kidx, qidx, qidx])
            h1_pos += np.einsum("pqrr->pq", eri_kkqq_pqrs) / nkpts

        one_body_mat[kidx] = hcore[kidx] + 0.5 * h1_neg + h1_pos
        lambda_one_body += np.sum(
            np.abs(one_body_mat[kidx].real) + np.abs(one_body_mat[kidx].imag))

    ############################################################################
    #
    # \lambda_{V} = \frac 12 \sum_{\Q}\sum_{n}^{M}\left
    # (\sum_{\K,pq}(|\Rea[L_{p\K,q\K-\Q,n}]| +
    #               |\Ima[L_{p\K,q\K-\Q,n}]|)\right)^{2}
    #
    # chol = [nkpts, nkpts, naux, nao, nao]
    #
    ############################################################################
    sf_obj.naux = old_naux  # reset naux to original value
    # this part needs to change
    lambda_two_body = 0.0
    for qidx in range(len(kpts)):
        # A and B are W
        A, B = sf_obj.build_AB_from_chol(qidx)  # [naux, nao * nk, nao * nk]
        A /= np.sqrt(nkpts)
        B /= np.sqrt(nkpts)
        # sum_q sum_n (sum_{pq} |Re{A_{pq}^n}| + |Im{A_{pq}^n|)^2
        lambda_two_body += np.sum(
            np.einsum("npq->n",
                      np.abs(A.real) + np.abs(A.imag))**2)
        lambda_two_body += np.sum(
            np.einsum("npq->n",
                      np.abs(B.real) + np.abs(B.imag))**2)

    lambda_two_body *= 0.5

    lambda_tot = lambda_one_body + lambda_two_body
    sf_data = SFHamiltonianProperties(
        lambda_total=lambda_tot,
        lambda_one_body=lambda_one_body,
        lambda_two_body=lambda_two_body,
        num_aux=sf_obj.naux,
    )
    return sf_data



#coverage:ignore
""" Drivers for various PySCF electronic structure routines """
from typing import Tuple, Optional
import sys
import h5py
import numpy as np
from pyscf import gto, scf, ao2mo, mcscf, lo, tools, cc
from pyscf.mcscf import avas


def stability(pyscf_mf):
    """
    Test wave function stability and re-optimize SCF.

    Args:
        pyscf_mf: PySCF mean field object (e.g. `scf.RHF()`)

    Returns:
        pyscf_mf: Updated PySCF mean field object
    """
    new_orbitals = pyscf_mf.stability()[0]
    new_1rdm = pyscf_mf.make_rdm1(new_orbitals, pyscf_mf.mo_occ)
    pyscf_mf = pyscf_mf.run(new_1rdm)

    return pyscf_mf


def localize(pyscf_mf, loc_type='pm', verbose=0):
    """ Localize orbitals given a PySCF mean-field object

    Args:
        pyscf_mf:  PySCF mean field object
        loc_type (str): localization type;
            Pipek-Mezey ('pm') or Edmiston-Rudenberg ('er')
        verbose (int): print level during localization

    Returns:
        pyscf_mf:  Updated PySCF mean field object with localized orbitals
    """
    # Note: After loading with `load_casfile_to_pyscf()` you can quiet message
    # by resetting mf.mol, i.e., mf.mol = gto.M(...)
    # but this assumes you have the *exact* molecular specification on hand.
    # I've gotten acceptable results by restoring mf.mol this way (usually
    # followed by calling mf.kernel()). But consistent localization is not a
    # given (not unique) despite restoring data this way, hence the message.
    if len(pyscf_mf.mol.atom) == 0:
        sys.exit("`localize()` requires atom loc. and atomic basis to be" + \
                 " defined.\n  " + \
                 "It also can be sensitive to the initial guess and MO" + \
                 " coefficients.\n  " + \
                 "Best to try re-creating the PySCF molecule and doing the" + \
                 " SCF, rather than\n  " + \
                 "try to load the mean-field object with" + \
                 " `load_casfile_to_pyscf()`. You can \n " + \
                 "try to provide the missing information, but consistency" + \
                 " cannot be guaranteed!")

    # Split-localize (localize DOCC, SOCC, and virtual separately)
    docc_idx = np.where(np.isclose(pyscf_mf.mo_occ, 2.))[0]
    socc_idx = np.where(np.isclose(pyscf_mf.mo_occ, 1.))[0]
    virt_idx = np.where(np.isclose(pyscf_mf.mo_occ, 0.))[0]

    # Pipek-Mezey
    if loc_type.lower() == 'pm':
        print("Localizing doubly occupied ... ", end="")
        loc_docc_mo = lo.PM(
            pyscf_mf.mol,
            pyscf_mf.mo_coeff[:, docc_idx]).kernel(verbose=verbose)
        print("singly occupied ... ", end="")
        loc_socc_mo = lo.PM(
            pyscf_mf.mol,
            pyscf_mf.mo_coeff[:, socc_idx]).kernel(verbose=verbose)
        print("virtual ... ", end="")
        loc_virt_mo = lo.PM(
            pyscf_mf.mol,
            pyscf_mf.mo_coeff[:, virt_idx]).kernel(verbose=verbose)
        print("DONE")

    # Edmiston-Rudenberg
    elif loc_type.lower() == 'er':
        print("Localizing doubly occupied ... ", end="")
        loc_docc_mo = lo.ER(
            pyscf_mf.mol,
            pyscf_mf.mo_coeff[:, docc_idx]).kernel(verbose=verbose)
        print("singly occupied ... ", end="")
        loc_socc_mo = lo.ER(
            pyscf_mf.mol,
            pyscf_mf.mo_coeff[:, socc_idx]).kernel(verbose=verbose)
        print("virtual ... ", end="")
        loc_virt_mo = lo.ER(
            pyscf_mf.mol,
            pyscf_mf.mo_coeff[:, virt_idx]).kernel(verbose=verbose)
        print("DONE")

    # overwrite orbitals with localized orbitals
    pyscf_mf.mo_coeff[:, docc_idx] = loc_docc_mo.copy()
    pyscf_mf.mo_coeff[:, socc_idx] = loc_socc_mo.copy()
    pyscf_mf.mo_coeff[:, virt_idx] = loc_virt_mo.copy()

    return pyscf_mf


def avas_active_space(pyscf_mf,
                      ao_list=None,
                      molden_fname='avas_localized_orbitals',
                      **kwargs):
    """ Return AVAS active space as PySCF molecule and mean-field object

    Args:
        pyscf_mf:  PySCF mean field object

    Kwargs:
        ao_list: list of strings of AOs (print mol.ao_labels() to see options)
                 Example: ao_list = ['H 1s', 'O 2p', 'O 2s'] for water
        verbose (bool): do additional print
        molden_fname (str): MOLDEN filename to save AVAS active space orbitals.
                            Default is to save
                            to 'avas_localized_orbitals.molden'
        **kwargs: other keyworded arguments to pass into avas.avas()

    Returns:
        pyscf_active_space_mol: Updated PySCF molecule object from
                                AVAS-selected active space
        pyscf_active_space_mf:  Updated PySCF mean field object from
                                AVAS-selected active space
    """

    # Note: requires openshell_option = 3 for this to work, which keeps all
    #     singly occupied in CAS
    # we also require canonicalize = False so that we don't destroy local orbs
    avas_output = avas.avas(pyscf_mf,
                            ao_list,
                            canonicalize=False,
                            openshell_option=3,
                            **kwargs)
    active_norb, active_ne, reordered_orbitals = avas_output

    active_alpha, _ = get_num_active_alpha_beta(pyscf_mf, active_ne)

    if molden_fname is not None:
        # save set of localized orbitals for active space
        if isinstance(pyscf_mf, scf.rohf.ROHF):
            frozen_alpha = pyscf_mf.nelec[0] - active_alpha
            assert frozen_alpha >= 0
        else:
            frozen_alpha = pyscf_mf.mol.nelectron // 2 - active_alpha
            assert frozen_alpha >= 0

        active_space_idx = slice(frozen_alpha, frozen_alpha + active_norb)
        active_mos = reordered_orbitals[:, active_space_idx]
        tools.molden.from_mo(pyscf_mf.mol,
                             molden_fname + '.molden',
                             mo_coeff=active_mos)

    # Choosing an active space changes the molecule ("freezing" electrons,
    # for example), so we
    # form the active space tensors first, then re-form the PySCF objects to
    # ensure consistency
    pyscf_active_space_mol, pyscf_active_space_mf = cas_to_pyscf(
        *pyscf_to_cas(pyscf_mf,
                      cas_orbitals=active_norb,
                      cas_electrons=active_ne,
                      avas_orbs=reordered_orbitals))

    return pyscf_active_space_mol, pyscf_active_space_mf


def cas_to_pyscf(h1, eri, ecore, num_alpha, num_beta):
    """ Return a PySCF molecule and mean-field object from pre-computed CAS Ham

    Args:
        h1 (ndarray) - 2D matrix containing one-body terms (MO basis)
        eri (ndarray) - 4D tensor containing two-body terms (MO basis)
        ecore (float) - frozen core electronic energy + nuclear repulsion energy
        num_alpha (int) - number of spin up electrons in CAS space
        num_beta (int) - number of spin down electrons in CAS space

    Returns:
        pyscf_mol: PySCF molecule object
        pyscf_mf:  PySCF mean field object
    """

    n_orb = len(h1)  # number orbitals
    assert [n_orb] * 4 == [*eri.shape]  # check dims are consistent

    pyscf_mol = gto.M()
    pyscf_mol.nelectron = num_alpha + num_beta
    n_orb = h1.shape[0]
    alpha_diag = [1] * num_alpha + [0] * (n_orb - num_alpha)
    beta_diag = [1] * num_beta + [0] * (n_orb - num_beta)

    # Assumes Hamiltonian is either RHF or ROHF ... should be OK since UHF will
    # have two h1s, etc.
    if num_alpha == num_beta:
        pyscf_mf = scf.RHF(pyscf_mol)
        scf_energy = ecore + \
                     2*np.einsum('ii',  h1[:num_alpha,:num_alpha]) + \
                     2*np.einsum('iijj',
                           eri[:num_alpha,:num_alpha,:num_alpha,:num_alpha]) - \
                       np.einsum('ijji',
                           eri[:num_alpha,:num_alpha,:num_alpha,:num_alpha])

    else:
        pyscf_mf = scf.ROHF(pyscf_mol)
        pyscf_mf.nelec = (num_alpha, num_beta)
        # grab singly and doubly occupied orbitals (assume high-spin open shell)
        docc = slice(None, min(num_alpha, num_beta))
        socc = slice(min(num_alpha, num_beta), max(num_alpha, num_beta))
        scf_energy = ecore + \
                     2.0*np.einsum('ii',h1[docc, docc]) + \
                         np.einsum('ii',h1[socc, socc]) + \
                     2.0*np.einsum('iijj',eri[docc, docc, docc, docc]) - \
                         np.einsum('ijji',eri[docc, docc, docc, docc]) + \
                         np.einsum('iijj',eri[socc, socc, docc, docc]) - \
                     0.5*np.einsum('ijji',eri[socc, docc, docc, socc]) + \
                         np.einsum('iijj',eri[docc, docc, socc, socc]) - \
                     0.5*np.einsum('ijji',eri[docc, socc, socc, docc]) + \
                     0.5*np.einsum('iijj',eri[socc, socc, socc, socc]) - \
                     0.5*np.einsum('ijji',eri[socc, socc, socc, socc])

    pyscf_mf.get_hcore = lambda *args: np.asarray(h1)
    pyscf_mf.get_ovlp = lambda *args: np.eye(h1.shape[0])
    pyscf_mf.energy_nuc = lambda *args: ecore
    pyscf_mf._eri = eri  # ao2mo.restore('8', np.zeros((8, 8, 8, 8)), 8)
    pyscf_mf.e_tot = scf_energy

    pyscf_mf.init_guess = '1e'
    pyscf_mf.mo_coeff = np.eye(n_orb)
    pyscf_mf.mo_occ = np.array(alpha_diag) + np.array(beta_diag)
    pyscf_mf.mo_energy, _ = np.linalg.eigh(pyscf_mf.get_fock())

    return pyscf_mol, pyscf_mf


def pyscf_to_cas(pyscf_mf,
                 cas_orbitals: Optional[int] = None,
                 cas_electrons: Optional[int] = None,
                 avas_orbs=None):
    """ Return CAS Hamiltonian tensors from a PySCF mean-field object

    Args:
        pyscf_mf: PySCF mean field object
        cas_orbitals (int, optional):  number of orbitals in CAS space,
                                       default all orbitals
        cas_electrons (int, optional): number of electrons in CAS space,
                                       default all electrons
        avas_orbs (ndarray, optional): orbitals selected by AVAS in PySCF

    Returns:
        h1 (ndarray) - 2D matrix containing one-body terms (MO basis)
        eri (ndarray) - 4D tensor containing two-body terms (MO basis)
        ecore (float) - frozen core electronic energy + nuclear repulsion energy
        num_alpha (int) - number of spin up electrons in CAS space
        num_beta (int) - number of spin down electrons in CAS space
    """

    # Only RHF or ROHF possible with mcscf.CASCI
    assert isinstance(pyscf_mf, scf.rhf.RHF)  # ROHF is child of RHF class

    if cas_orbitals is None:
        cas_orbitals = len(pyscf_mf.mo_coeff)
    if cas_electrons is None:
        cas_electrons = pyscf_mf.mol.nelectron

    cas = mcscf.CASCI(pyscf_mf, ncas=cas_orbitals, nelecas=cas_electrons)
    h1, ecore = cas.get_h1eff(mo_coeff=avas_orbs)
    eri = cas.get_h2cas(mo_coeff=avas_orbs)
    eri = ao2mo.restore('s1', eri, h1.shape[0])  # chemist convention (11|22)
    ecore = float(ecore)

    num_alpha, num_beta = get_num_active_alpha_beta(pyscf_mf, cas_electrons)

    return h1, eri, ecore, num_alpha, num_beta


def get_num_active_alpha_beta(pyscf_mf, cas_electrons):
    """ Return number of alpha and beta electrons in the active space given
        number of CAS electrons
        This assumes that all the unpaired electrons are in the active space

    Args:
        pyscf_mf: PySCF mean field object
        cas_orbitals (int):  number of electrons in CAS space,

    Returns:
        num_alpha (int): number of alpha (spin-up) electrons in active space
        num_beta (int):  number of beta (spin-down) electrons in active space
    """
    # Sanity checks and active space info
    total_electrons = pyscf_mf.mol.nelectron
    frozen_electrons = total_electrons - cas_electrons
    assert frozen_electrons % 2 == 0

    # ROHF == RHF but RHF != ROHF, and we only do either RHF or ROHF
    if isinstance(pyscf_mf, scf.rohf.ROHF):
        frozen_alpha = frozen_electrons // 2
        frozen_beta = frozen_electrons // 2
        num_alpha = pyscf_mf.nelec[0] - frozen_alpha
        num_beta = pyscf_mf.nelec[1] - frozen_beta
        assert np.isclose(num_beta + num_alpha, cas_electrons)

    else:
        assert cas_electrons % 2 == 0
        num_alpha = cas_electrons // 2
        num_beta = cas_electrons // 2

    return num_alpha, num_beta


def load_casfile_to_pyscf(fname,
                          num_alpha: Optional[int] = None,
                          num_beta: Optional[int] = None):
    """ Load CAS Hamiltonian from pre-computed HD5 file into a PySCF molecule
        and mean-field object

    Args:
        fname (str): path to hd5 file to be created containing CAS one and two
                     body terms
        num_alpha (int, optional): number of spin up electrons in CAS space
        num_beta (int, optional):  number of spin down electrons in CAS space

    Returns:
        pyscf_mol: PySCF molecule object
        pyscf_mf:  PySCF mean field object
    """

    with h5py.File(fname, "r") as f:
        eri = np.asarray(f['eri'][()])
        # h1 one body elements are sometimes called different things. Try a few.
        try:
            h1 = np.asarray(f['h0'][()])
        except KeyError:
            try:
                h1 = np.asarray(f['hcore'][()])
            except KeyError:
                try:
                    h1 = np.asarray(f['h1'][()])
                except KeyError:
                    raise KeyError("Could not find 1-electron Hamiltonian")
        # ecore sometimes exists, and sometimes as enuc (no frozen electrons)
        try:
            ecore = float(f['ecore'][()])
        except KeyError:
            try:
                ecore = float(f['enuc'][()])
            except KeyError:
                ecore = 0.0
        # read the number of spin up and spin down electrons if not input
        if (num_alpha is None) or (num_beta is None):
            try:
                num_alpha = int(f['active_nalpha'][()])
            except KeyError:
                sys.exit("In `load_casfile_to_pyscf()`: \n" + \
                         " No values found on file for num_alpha " + \
                         "(key: 'active_nalpha' in h5). " + \
                         " Try passing in a value for num_alpha, or" + \
                         " re-check integral file.")
            try:
                num_beta = int(f['active_nbeta'][()])
            except KeyError:
                sys.exit("In `load_casfile_to_pyscf()`: \n" + \
                         " No values found on file for num_beta " + \
                         "(key: 'active_nbeta' in h5). " + \
                         " Try passing in a value for num_beta, or" + \
                         " re-check integral file.")

    pyscf_mol, pyscf_mf = cas_to_pyscf(h1, eri, ecore, num_alpha, num_beta)

    return pyscf_mol, pyscf_mf


def save_pyscf_to_casfile(fname,
                          pyscf_mf,
                          cas_orbitals: Optional[int] = None,
                          cas_electrons: Optional[int] = None,
                          avas_orbs=None):
    """ Save CAS Hamiltonian from a PySCF mean-field object to an HD5 file

    Args:
        fname (str): path to hd5 file to be created containing CAS terms
        pyscf_mf: PySCF mean field object
        cas_orbitals (int, optional):  number of orb in CAS space, default all
        cas_electrons (int, optional): number of elec in CAS, default all elec
        avas_orbs (ndarray, optional): orbitals selected by AVAS in PySCF
    """
    h1, eri, ecore, num_alpha, num_beta = \
        pyscf_to_cas(pyscf_mf, cas_orbitals, cas_electrons, avas_orbs)

    with h5py.File(fname, 'w') as fid:
        fid.create_dataset('ecore', data=float(ecore), dtype=float)
        fid.create_dataset(
            'h0',
            data=h1)  # note the name change to be consistent with THC paper
        fid.create_dataset('eri', data=eri)
        fid.create_dataset('active_nalpha', data=int(num_alpha), dtype=int)
        fid.create_dataset('active_nbeta', data=int(num_beta), dtype=int)


def factorized_ccsd_t(pyscf_mf, eri_rr = None, use_kernel = True,\
    no_triples=False) -> Tuple[float, float, float]:
    """ Compute CCSD(T) energy using rank-reduced ERIs

    Args:
        pyscf_mf - PySCF mean field object
        eri_rr (ndarray) - rank-reduced ERIs, or use full ERIs from pyscf_mf
        use_kernel (bool) - re-do SCF, using canonical orbitals for one-body?
        no_triples (bool) - skip the perturbative triples correction? (CCSD)

    Returns:
        e_scf (float) - SCF energy
        e_cor (float) - Correlation energy from CCSD(T)
        e_tot (float) - Total energy; i.e. SCF + Corr energy from CCSD(T)
    """
    h1, eri_full, ecore, num_alpha, num_beta = pyscf_to_cas(pyscf_mf)

    # If no rank-reduced ERIs, use the full (possibly local) ERIs from pyscf_mf
    if eri_rr is None:
        eri_rr = eri_full

    e_scf, e_cor, e_tot = ccsd_t(h1, eri_rr, ecore, num_alpha, num_beta,\
        eri_full, use_kernel, no_triples)

    return e_scf, e_cor, e_tot


def ccsd_t(h1, eri, ecore, num_alpha: int, num_beta: int, eri_full = None,\
    use_kernel=True, no_triples=False) -> Tuple[float, float, float]:
    """ Helper function to do CCSD(T) on set of one- and two-body Hamil elems

    Args:
        h1 (ndarray) - 2D matrix containing one-body terms (MO basis)
        eri (ndarray) - 4D tensor containing two-body terms (MO basis)
                        may be from integral factorization (e.g. SF/DF/THC)
        ecore (float) - frozen core electronic energy + nuclear repulsion energy
        num_alpha (int) - number of spin alpha electrons in Hamiltonian
        num_beta (int) - number of spin beta electrons in Hamiltonian
        eri_full (ndarray) - optional 4D tensor containing full two-body
            terms (MO basis) for the SCF procedure only
        use_kernel (bool) - re-run SCF prior to doing CCSD(T)?
        no_triples (bool) - skip the perturbative triples correction? (CCSD)

    Returns:
        e_scf (float) - SCF energy
        e_cor (float) - Correlation energy from CCSD(T)
        e_tot (float) - Total energy; i.e. SCF + Corr energy from CCSD(T)
    """

    mol = gto.M()
    mol.nelectron = num_alpha + num_beta
    n_orb = h1.shape[0]
    alpha_diag = [1] * num_alpha + [0] * (n_orb - num_alpha)
    beta_diag = [1] * num_beta + [0] * (n_orb - num_beta)

    # If eri_full not provided, use (possibly rank-reduced) ERIs for check
    if eri_full is None:
        eri_full = eri

    # either RHF or ROHF ... should be OK since UHF will have two h1s, etc.
    if num_alpha == num_beta:
        mf = scf.RHF(mol)
        scf_energy = ecore + \
                     2*np.einsum('ii',h1[:num_alpha,:num_alpha]) + \
                     2*np.einsum('iijj',eri_full[:num_alpha,\
                                                 :num_alpha,\
                                                 :num_alpha,\
                                                 :num_alpha]) - \
                       np.einsum('ijji',eri_full[:num_alpha,\
                                                 :num_alpha,\
                                                 :num_alpha,\
                                                 :num_alpha])

    else:
        mf = scf.ROHF(mol)
        mf.nelec = (num_alpha, num_beta)
        # grab singly and doubly occupied orbitals (assume high-spin open shell)
        docc = slice(None, min(num_alpha, num_beta))
        socc = slice(min(num_alpha, num_beta), max(num_alpha, num_beta))
        scf_energy = ecore + \
                     2.0*np.einsum('ii',h1[docc, docc]) + \
                         np.einsum('ii',h1[socc, socc]) + \
                     2.0*np.einsum('iijj',eri_full[docc, docc, docc, docc]) - \
                         np.einsum('ijji',eri_full[docc, docc, docc, docc]) + \
                         np.einsum('iijj',eri_full[socc, socc, docc, docc]) - \
                     0.5*np.einsum('ijji',eri_full[socc, docc, docc, socc]) + \
                         np.einsum('iijj',eri_full[docc, docc, socc, socc]) - \
                     0.5*np.einsum('ijji',eri_full[docc, socc, socc, docc]) + \
                     0.5*np.einsum('iijj',eri_full[socc, socc, socc, socc]) - \
                     0.5*np.einsum('ijji',eri_full[socc, socc, socc, socc])

    mf.get_hcore = lambda *args: np.asarray(h1)
    mf.get_ovlp = lambda *args: np.eye(h1.shape[0])
    mf.energy_nuc = lambda *args: ecore
    mf._eri = eri_full  # ao2mo.restore('8', np.zeros((8, 8, 8, 8)), 8)

    mf.init_guess = '1e'
    mf.mo_coeff = np.eye(n_orb)
    mf.mo_occ = np.array(alpha_diag) + np.array(beta_diag)
    w, _ = np.linalg.eigh(mf.get_fock())
    mf.mo_energy = w

    # Rotate the interaction tensors into the canonical basis.
    # Reiher and Li tensors, for example, are read-in in the local MO basis,
    # which is not optimal for the CCSD(T) calculation (canonical gives better
    # energy estimate whereas QPE is invariant to choice of basis)
    if use_kernel:
        mf.conv_tol = 1e-7
        mf.init_guess = '1e'
        mf.verbose = 4
        mf.diis_space = 24
        mf.level_shift = 0.5
        mf.conv_check = False
        mf.max_cycle = 800
        mf.kernel(mf.make_rdm1(mf.mo_coeff,
                               mf.mo_occ))  # use MO info to generate guess
        mf = stability(mf)
        mf = stability(mf)
        mf = stability(mf)

        # Check if SCF has changed by doing restart, and print warning if so
        try:
            assert np.isclose(scf_energy, mf.e_tot, rtol=1e-14)
        except AssertionError:
            print(
                "WARNING: E(SCF) from input integrals does not match E(SCF)" + \
                " from mf.kernel()")
            print("  Will use E(SCF) = {:12.6f} from mf.kernel going forward.".
                  format(mf.e_tot))
        print("E(SCF, ints) = {:12.6f} whereas E(SCF) = {:12.6f}".format(
            scf_energy, mf.e_tot))

        # New SCF energy and orbitals for CCSD(T)
        scf_energy = mf.e_tot

    # Now re-set the eri's to the (possibly rank-reduced) ERIs
    mf._eri = eri
    mf.mol.incore_anyway = True

    mycc = cc.CCSD(mf)
    mycc.max_cycle = 800
    mycc.conv_tol = 1E-8
    mycc.conv_tol_normt = 1E-4
    mycc.diis_space = 24
    mycc.verbose = 4
    mycc.kernel()

    if no_triples:
        et = 0.0
    else:
        et = mycc.ccsd_t()

    e_scf = scf_energy  # may be read-in value or 'fresh' SCF value
    e_cor = mycc.e_corr + et
    e_tot = e_scf + e_cor

    print("E(SCF):       ", e_scf)
    print("E(cor):       ", e_cor)
    print("Total energy: ", e_tot)
    return e_scf, e_cor, e_tot


def open_shell_t1_d1(t1a, t1b, mo_occ, nalpha, nbeta):
    """
    T1-diagnostic for open-shell is defined w.r.t Sx eigenfunction of T1
        where reference is ROHF.

    given i double occ, c unoccupied, x is single occuplied The T1 amps
      (high spin) in Sz basis are:
    T1 = t_{ia}^{ca}(ca^ ia) + t_{ib}^{cb}(cb^ ib)
       + t_{xa}^{ca}(ca^ xa) + t_{ib}^{xb}(xb^ ib)
    T1 in the Sx basis are
    T1 = f_{i}^{c}E_{ci} + v_{i}^{c}A_{ci}
       + sqrt(2)f_{x}^{c}(ca^ xa) + sqrt(2)f_{i}^{x}(xb^ ib)

    where E_{ci} = ca^ ia + cb^ ib and A_{ci} = ca^ ia - cb^ ib.

    See:  The Journal of Chemical Physics 98, 9734 (1993);
              doi: 10.1063/1.464352
          Chemical Physics Letters 372 (2003) 362–367;
              doi:10.1016/S0009-2614(03)00435-4

    based on these and two papers from Lee the T1-openshell diagnostic is

    sqrt(sum_{ia}(f_{ia})^2 + 2sum_{xa}(t_{xa}^{ca})^2
        + 2 sum_{ix}(t_{ib}^{xb})^2) / 2 sqrt{N}

    To get this relate eqs 3-7 from Chemical Physics Letters 372 (2003) 362–367
    to Eqs. 45, 46, and 51 from Journal of Chemical Physics 98, 9734 (1993);
    doi: 10.1063/1.464352.
    """
    # compute t1-diagnostic
    docc_idx = np.where(np.isclose(mo_occ, 2.))[0]
    socc_idx = np.where(np.isclose(mo_occ, 1.))[0]
    virt_idx = np.where(np.isclose(mo_occ, 0.))[0]
    t1a_docc = t1a[docc_idx, :]  # double occ-> virtual
    t1b_docc = t1b[docc_idx, :][:, -len(virt_idx):]  # double occ-> virtual
    if len(socc_idx) > 0:
        t1_xa = t1a[socc_idx, :]  # single occ -> virtual
        t1_ix = t1b[docc_idx, :][:, :len(socc_idx)]  # double occ -> single occ
    else:
        t1_xa = np.array(())
        t1_ix = np.array(())

    if nalpha - nbeta + len(virt_idx) != t1b.shape[1]:
        raise ValueError(
            "Inconsistent shapes na {}, nb {}, t1b.shape {},{}".format(
                nalpha, nbeta, t1b.shape[0], t1b.shape[1]))

    if t1a_docc.shape != (len(docc_idx), len(virt_idx)):
        raise ValueError("T1a_ia does not have the right shape")
    if t1b_docc.shape != (len(docc_idx), len(virt_idx)):
        raise ValueError("T1b_ia does not have the right shape")
    if len(socc_idx) > 0:
        if t1_ix.shape != (len(docc_idx), len(socc_idx)):
            raise ValueError("T1_ix does not have the right shape")
        if t1_xa.shape != (len(socc_idx), len(virt_idx)):
            raise ValueError("T1_xa does not have the right shape")

    t1_diagnostic = np.sqrt(
        np.sum((t1a_docc + t1b_docc)**2) + 2 * np.sum(t1_xa**2) +
        2 * np.sum(t1_ix**2)) / (2 * np.sqrt(nalpha + nbeta))
    # compute D1-diagnostic
    f_ia = 0.5 * (t1a_docc + t1b_docc)
    s_f_ia_2, _ = np.linalg.eigh(f_ia @ f_ia.T)
    s_f_ia_2_norm = np.sqrt(np.max(s_f_ia_2, initial=0))

    if len(socc_idx) > 0:
        f_xa = np.sqrt(1 / 2) * t1_xa
        f_ix = np.sqrt(1 / 2) * t1_ix
        s_f_xa_2, _ = np.linalg.eigh(f_xa @ f_xa.T)
        s_f_ix_2, _ = np.linalg.eigh(f_ix @ f_ix.T)
    else:
        s_f_xa_2 = np.array(())
        s_f_ix_2 = np.array(())
    s_f_xa_2_norm = np.sqrt(np.max(s_f_xa_2, initial=0))
    s_f_ix_2_norm = np.sqrt(np.max(s_f_ix_2, initial=0))

    d1_diagnostic = np.max(
        np.array([s_f_ia_2_norm, s_f_xa_2_norm, s_f_ix_2_norm]))

    return t1_diagnostic, d1_diagnostic



#coverage:ignore
"""Test cases for compute_lambda_df.py
"""
from os import path

import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES, df

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from openfermion.resource_estimates.molecule import load_casfile_to_pyscf


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason="pyscf and/or jax not installed.")
def test_reiher_df_lambda():
    """ Reproduce Reiher et al orbital DF lambda from paper """

    THRESH = 0.00125
    NAME = path.join(path.dirname(__file__), '../integrals/eri_reiher.h5')
    _, mf = load_casfile_to_pyscf(NAME, num_alpha=27, num_beta=27)
    eri_rr, LR, L, Lxi = df.factorize(mf._eri, thresh=THRESH)
    total_lambda = df.compute_lambda(mf, LR)
    assert eri_rr.shape[0] * 2 == 108
    assert L == 360
    assert Lxi == 13031
    assert np.isclose(np.round(total_lambda, decimals=1), 294.8)



#coverage:ignore
""" Determine costs for DF decomposition in QC """
from typing import Tuple
import numpy as np
from numpy.lib.scimath import arccos, arcsin  # has analytc continuation to cplx
from openfermion.resource_estimates.utils import QR, QI, power_two


def compute_cost(n: int,
                 lam: float,
                 dE: float,
                 L: int,
                 Lxi: int,
                 chi: int,
                 beta: int,
                 stps: int,
                 verbose: bool = False) -> Tuple[int, int, int]:
    """ Determine fault-tolerant costs using DF decomposition in quantum chem

    Args:
        n (int) - the number of spin-orbitals
        lam (float) - the lambda-value for the Hamiltonian
        dE (float) - allowable error in phase estimation
        L (int) - the rank of the first decomposition
        Lxi (int) - the total number of eigenvectors
        chi (int) - equivalent to aleph_1 and aleph_2 in the document, the
            number of bits for the representation of the coefficients
        beta (int) - equivalent to beth in the document, the number of bits
            for the rotations
        stps (int) - an approximate number of steps to choose the precision of
            single qubit rotations in preparation of the equal superpositn state
        verbose (bool) - do additional printing of intermediates?

    Returns:
        step_cost (int) - Toffolis per step
        total_cost (int) - Total number of Toffolis
        ancilla_cost (int) - Total ancilla cost
    """

    # The number of bits used for the second register.
    nxi = np.ceil(np.log2(n // 2))

    # The number of bits for the contiguous register.
    nLxi = np.ceil(np.log2(Lxi + n // 2))

    # The number of bits used for the first register.
    nL = np.ceil(np.log2(L + 1))

    # The power of 2 that is a factor of L + 1
    eta = power_two(L + 1)

    oh = [0] * 20
    for p in range(20):
        # JJG note: arccos arg may be > 1
        v = np.round(np.power(2,p+1) / (2 * np.pi) * arccos(np.power(2,nL) /\
            np.sqrt((L + 1)/2**eta)/2))
        oh[p] = np.real(stps * (1 / (np.sin(3 * arcsin(np.cos(v * 2 * np.pi / \
            np.power(2,p+1)) * \
            np.sqrt((L + 1)/2**eta) / np.power(2,nL)))**2) - 1) + 4 * (p + 1))

    # Bits of precision for rotation
    br = int(np.argmin(oh) + 1)

    # The following costs are from the list starting on page 50.

    # The cost for preparing an equal superposition for preparing the first
    # register in step 1 (a). We double this cost to account for the inverse.
    cost1a = 2 * (3 * nL + 2 * br - 3 * eta - 9)

    # The output size for the QROM for the first state preparation in Eq. (C27)
    bp1 = nL + chi

    # The cost of the QROM for the first state preparation in step 1 (b) and
    # its inverse.
    cost1b = QR(L + 1, bp1)[1] + QI(L + 1)[1]

    # The cost for the inequality test, controlled swap and their inverse in
    # steps 1 (c) and (d)
    cost1cd = 2 * (chi + nL)

    # The total cost for preparing the first register in step 1.
    cost1 = cost1a + cost1b + cost1cd

    # The output size for the QROM for the data to prepare the equal
    # superposition on the second register, as given in Eq. (C29).
    bo = nxi + nLxi + br + 1

    # This is step 2. This is the cost of outputting the data to prepare the
    # equal superposition on the second register. We will assume it is not
    # uncomputed, because we want to keep the offset for applying the QROM for
    # outputting the rotations.
    cost2 = QR(L + 1, bo)[1] + QI(L + 1)[1]

    # The number of bits for rotating the ancilla for the second preparation.
    # We are just entering this manually because it is a typical value.
    br = 7

    # The cost of preparing an equal superposition over the second register in
    # a controlled way. We pay this cost 4 times.
    cost3a = 4 * (7 * nxi + 2 * br - 6)

    # The cost of the offset to apply the QROM for state preparation on the
    # second register.
    cost3b = 4 * (nLxi - 1)

    bp2 = nxi + chi + 2

    # The cost of the QROMs and inverse QROMs for the state preparation, where
    # in the first one we need + n/2 to account for the one-electron terms.
    cost3c = QR(Lxi + n // 2, bp2)[1] + QI(Lxi + n // 2)[1] + QR(
        Lxi, bp2)[1] + QI(Lxi)[1]

    # The inequality test and state preparations.
    cost3d = 4 * (nxi + chi)

    # The total costs for state preparations on register 2.
    cost3 = cost3a + cost3b + cost3c + cost3d

    # The cost of adding offsets in steps 4 (a) and (h).
    cost4ah = 4 * (nLxi - 1)

    # The costs of the QROMs and their inverses in steps 4 (b) and (g).
    cost4bg = QR(Lxi + n // 2, n * beta // 2)[1] + QI(Lxi + n // 2)[1] + QR(
        Lxi, n * beta // 2)[1] + QI(Lxi)[1]

    # The cost of the controlled swaps based on the spin qubit in steps 4c and f
    cost4cf = 2 * n

    # The controlled rotations in steps 4 (d) and (f).
    cost4df = 4 * n * (beta - 2)

    # The controlled Z operations in the middle for step 4 (e).
    cost4e = 3

    # This is the cost of the controlled rotations for step 4.
    cost4 = cost4ah + cost4bg + cost4cf + cost4df + cost4e

    # This is the cost of the reflection on the second register from step 6.
    cost6 = nxi + chi + 2

    # The cost of the final reflection req'd to construct the step of the
    # quantum walk from step 9.
    cost9 = nL + nxi + chi + 1

    # The extra two qubits for unary iteration and making the rflxn controlled.
    cost10 = 2

    # The Toffoli cost for a single step
    cost = cost1 + cost2 + cost3 + cost4 + cost6 + cost9 + cost10

    # The number of steps needed
    iters = np.ceil(np.pi * lam / (2 * dE))

    # Now the number of qubits from the list on page 54.

    k1 = np.power(2, QR(Lxi + n // 2, n * beta // 2)[0])

    # The control register for phase estimation and iteration on it.
    ac1 = np.ceil(np.log2(iters + 1)) * 2 - 1

    # The system qubits
    ac2 = n

    # The first register prepared, a rotated qubit and a flag qubit.
    ac3 = nL + 2

    # The output of the QROM, the equal superposition state and a flag qubit.
    ac4 = nL + chi * 2 + 1

    # The data used for preparing the equal superposition state on the 2nd reg
    ac5 = bo

    # The second register, a rotated qubit and a flag qubit.
    ac6 = nxi + 2

    # The second preparation QROM output.
    ac8 = bp2

    # The equal superposition state and the result of the inequality test.
    ac9 = chi + 1

    # The angles for rotations.
    ac10 = k1 * n * beta // 2

    # The phase gradient state.
    ac11 = beta

    # A control qubit for the spin.
    ac12 = 1

    # A T state.
    ac13 = 1

    if verbose:
        print("[*] Top of routine")
        print("  [+] nxi = ", nxi)
        print("  [+] nLxi = ", nLxi)
        print("  [+] nL = ", nL)
        print("  [+] eta = ", eta)
        print("  [+] cost3 = ", cost3)
        print("  [+] cost4 = ", cost4)
        print("  [+] cost = ", cost)
        print("  [+] iters = ", iters)

    ancilla_cost = ac1 + ac2 + ac3 + ac4 + ac5 + ac6 + ac8 + ac9 + ac10 + ac11\
                 + ac12 + ac13

    # Sanity checks before returning as int
    assert cost.is_integer()
    assert iters.is_integer()
    assert ancilla_cost.is_integer()

    step_cost = int(cost)
    total_cost = int(cost * iters)
    ancilla_cost = int(ancilla_cost)

    return step_cost, total_cost, ancilla_cost



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Union, Tuple
import itertools
import numpy as np
import numpy.typing as npt

from pyscf.pbc import scf
from pyscf.pbc.lib.kpts_helper import KptsHelper, loop_kkk

from openfermion.resource_estimates.pbc.hamiltonian import (
    build_momentum_transfer_mapping,)


def _symmetric_two_body_terms(quad: Tuple[int, ...], complex_valued: bool):
    p, q, r, s = quad
    yield p, q, r, s
    yield q, p, s, r
    yield s, r, q, p
    yield r, s, p, q
    if not complex_valued:
        yield p, s, r, q
        yield q, r, s, p
        yield s, p, q, r
        yield r, q, p, s


def unique_iter(nmo: int):
    """Iterate over unique pqrs indices"""
    seen = set()
    for quad in itertools.product(range(nmo), repeat=4):
        if quad not in seen:
            seen |= set(_symmetric_two_body_terms(quad, True))
            yield tuple(quad)


def _pq_rs_two_body_terms(quad):
    """Symmetry inequivalent indices when kp = kq and kr = ks.

    A subset of the four-fold symmetry can be applied
    (pkp, qkp|rkr, skr) = (qkp,pkp|skr,rkr) by complex conjucation
    """
    p, q, r, s = quad
    yield p, q, r, s
    yield q, p, s, r


def unique_iter_pq_rs(nmo: int):
    """Iterate over unique symmetry pqrs indices"""
    seen = set()
    for quad in itertools.product(range(nmo), repeat=4):
        if quad not in seen:
            seen |= set(_pq_rs_two_body_terms(quad))
            yield tuple(quad)


def _ps_qr_two_body_terms(quad):
    """Symmetry inequivalent indices when kp = ks and kq = kr.

    A subset of the four-fold symmetry can be applied (pkp,qkq|rkq,skp) ->
    (skp,rkq|qkq,pkp) by complex conj and dummy index exchange
    """
    p, q, r, s = quad
    yield p, q, r, s
    yield s, r, q, p


def unique_iter_ps_qr(nmo):
    """Iterate over unique ps qr indices"""
    seen = set()
    for quad in itertools.product(range(nmo), repeat=4):
        if quad not in seen:
            seen |= set(_ps_qr_two_body_terms(quad))
            yield tuple(quad)


def _pr_qs_two_body_terms(quad):
    """Symmetry inequivalent indices when kp = kr and kq = ks.

    A subset of the four-fold symmetry can be applied (pkp,qkq|rkp,skq) ->
    (rkp,skq|pkp,rkq) by dummy index exchange
    """
    p, q, r, s = quad
    yield p, q, r, s
    yield r, s, p, q


def unique_iter_pr_qs(nmo):
    """Iterate over unique pr qs indices"""
    seen = set()
    for quad in itertools.product(range(nmo), repeat=4):
        if quad not in seen:
            seen |= set(_pr_qs_two_body_terms(quad))
            yield tuple(quad)


class SparseFactorization:

    def __init__(self,
                 cholesky_factor: np.ndarray,
                 kmf: scf.HF,
                 threshold=1.0e-14):
        """Initialize a ERI object for CCSD from sparse integrals.

        Args:
            cholesky_factor: Cholesky factor tensor that is
                [nkpts, nkpts, naux, nao, nao].
                To see how to generate this see pyscf_chol_from_df
            kmf: pyscf k-object.  Currently only used to obtain the number of
                k-points.  Must have an attribute kpts which len(self.kmf.kpts)
                returns number of kpts.
            threshold: Default 1.0E-8 is the value at which to ignore the
                integral
        """
        self.chol = cholesky_factor
        self.kmf = kmf
        self.nk = len(self.kmf.kpts)
        self.nao = cholesky_factor[0, 0].shape[-1]
        k_transfer_map = build_momentum_transfer_mapping(
            self.kmf.cell, self.kmf.kpts)
        self.k_transfer_map = k_transfer_map
        self.threshold = threshold

    def get_total_unique_terms_above_thresh(self, return_nk_counter=False
                                           ) -> Union[int, Tuple[int, int]]:
        """Determine all unique (pkp, qkq|rkr, sks).

        Accounts for momentum conservation and four fold symmetry.

        Args:
            return_nk_counter: Return number of visited k-points.

        Returns:
            num_sym_unique: Number of symmetry unique terms above the threshold
        """
        kpts_helper = KptsHelper(self.kmf.cell, self.kmf.kpts)
        nkpts = len(self.kmf.kpts)
        completed = np.zeros((nkpts, nkpts, nkpts), dtype=bool)
        counter = 0
        nk_counter = 0
        for kvals in loop_kkk(nkpts):
            kp, kq, kr = kvals
            ks = kpts_helper.kconserv[kp, kq, kr]
            if not completed[kp, kq, kr]:
                nk_counter += 1
                eri_block = self.get_eri([kp, kq, kr, ks])
                if kp == kq == kr == ks:
                    completed[kp, kq, kr] = True
                    n = self.nao
                    assert all(nx == n for nx in eri_block.shape)
                    Dd = np.zeros((n,), dtype=eri_block.dtype)
                    for p in range(n):
                        Dd[p] = eri_block[p, p, p, p]
                        eri_block[p, p, p, p] = 0.0
                    Dp = np.zeros((n, n), dtype=eri_block.dtype)
                    for p, q in itertools.product(range(n), repeat=2):
                        Dp[p, q] = eri_block[p, q, p, q]
                        eri_block[p, q, p, q] = 0.0
                    Dc = np.zeros((n, n), dtype=eri_block.dtype)
                    for p, r in itertools.product(range(n), repeat=2):
                        Dc[p, r] = eri_block[p, p, r, r]
                        eri_block[p, p, r, r] = 0.0
                    Dpc = np.zeros((n, n), dtype=eri_block.dtype)
                    for p, q in itertools.product(range(n), repeat=2):
                        Dpc[p, q] = eri_block[p, q, q, p]
                        eri_block[p, q, q, p] = 0.0
                    counter += np.count_nonzero(Dd)
                    counter += np.count_nonzero(Dp) // 2
                    counter += np.count_nonzero(Dc) // 2
                    counter += np.count_nonzero(Dpc) // 2
                    counter += np.count_nonzero(eri_block) // 4
                elif kp == kq and kr == ks:
                    completed[kp, kq, kr] = True
                    completed[kr, ks, kp] = True
                    n = self.nao
                    assert all(nx == n for nx in eri_block.shape)
                    Dc = np.zeros((n, n), dtype=eri_block.dtype)
                    for p, r in itertools.product(range(n), repeat=2):
                        Dc[p, r] = eri_block[p, p, r, r]
                        eri_block[p, p, r, r] = 0.0
                    counter += np.count_nonzero(Dc)
                    counter += np.count_nonzero(eri_block) // 2
                elif kp == ks and kq == kr:
                    completed[kp, kq, kr] = True
                    completed[kr, ks, kp] = True
                    n = self.nao
                    assert all(nx == n for nx in eri_block.shape)
                    Dpc = np.zeros((n, n), dtype=eri_block.dtype)
                    for p, q in itertools.product(range(n), repeat=2):
                        Dpc[p, q] = eri_block[p, q, q, p]
                        eri_block[p, q, q, p] = 0.0
                    counter += np.count_nonzero(Dpc)
                    counter += np.count_nonzero(eri_block) // 2
                elif kp == kr and kq == ks:
                    completed[kp, kq, kr] = True
                    completed[kq, kp, ks] = True
                    n = self.nao
                    assert all(nx == n for nx in eri_block.shape)
                    Dp = np.zeros((n, n), dtype=eri_block.dtype)
                    for p, q in itertools.product(range(n), repeat=2):
                        Dp[p, q] = eri_block[p, q, p, q]
                        eri_block[p, q, p, q] = 0.0
                    counter += np.count_nonzero(Dp)
                    counter += np.count_nonzero(eri_block) // 2
                else:
                    counter += np.count_nonzero(eri_block)
                    completed[kp, kq, kr] = True
                    completed[kr, ks, kp] = True
                    completed[kq, kp, ks] = True
                    completed[ks, kr, kq] = True
        if return_nk_counter:
            return counter, nk_counter
        return counter

    def get_eri(self, ikpts, check_eq=False) -> npt.NDArray:
        """Construct (pkp qkq| rkr sks) via  cholesky factors.

        Note: 3-tensor L_{sks, rkr} = L_{rkr, sks}^{*}

        Args:
            ikpts: list of four integers representing the index of the kpoint in
                self.kmf.kpts
            check_eq: optional value to confirm a symmetry in the Cholesky
                vectors. (Default value = False)

        Returns:
            eris: ([pkp][qkq]|[rkr][sks])
        """
        ikp, ikq, ikr, iks = ikpts
        if check_eq:
            assert np.allclose(
                np.einsum(
                    "npq,nsr->pqrs",
                    self.chol[ikp, ikq],
                    self.chol[iks, ikr].conj(),
                    optimize=True,
                ),
                np.einsum(
                    "npq,nrs->pqrs",
                    self.chol[ikp, ikq],
                    self.chol[ikr, iks],
                    optimize=True,
                ),
            )
        eri = np.einsum(
            "npq,nsr->pqrs",
            self.chol[ikp, ikq],
            self.chol[iks, ikr].conj(),
            optimize=True,
        )
        zero_mask = np.abs(eri) < self.threshold
        eri[zero_mask] = 0
        return eri

    def get_eri_exact(self, kpts) -> npt.NDArray:
        """Construct (pkp qkq| rkr sks) exactly from mean-field object.

        This is for constructing the J and K like terms needed for the
        one-body component lambda.

        Args:
            kpts: list of four integers representing the index of the kpoint
                in self.kmf.kpts

        Returns:
            eris: ([pkp][qkq]|[rkr][sks])
        """
        ikp, ikq, ikr, iks = kpts
        return np.einsum(
            "npq,nsr->pqrs",
            self.chol[ikp, ikq],
            self.chol[iks, ikr].conj(),
            optimize=True,
        )



"""Test cases for costing_sparse.py
"""
from openfermion.resource_estimates.sparse.costing_sparse import cost_sparse


def test_reiher_sparse():
    """ Reproduce Reiher et al orbital sparse FT costs from paper """
    DE = 0.001
    CHI = 10

    # Reiher et al orbitals
    N = 108
    LAM = 2135.3
    D = 705831

    # Here we're using an initial calculation with a very rough estimate of the
    # number of steps to give a more accurate number of steps. Then we input
    # that into the function again.
    output = cost_sparse(N, LAM, D, DE, CHI, stps=20000)
    stps1 = output[0]
    output = cost_sparse(N, LAM, D, DE, CHI, stps1)
    assert output == (26347, 88371052334, 2190)


def test_li_sparse():
    """ Reproduce Li et al orbital sparse FT costs from paper """
    DE = 0.001
    CHI = 10

    # Li et al orbitals
    N = 152
    LAM = 1547.3
    D = 440501

    # Here we're using an initial calculation with a very rough estimate of the
    # number of steps to give a more accurate number of steps. Then we input
    # that into the function again.
    output = cost_sparse(N, LAM, D, DE, CHI, stps=20000)
    stps2 = output[0]
    output = cost_sparse(N, LAM, D, DE, CHI, stps2)
    assert output == (18143, 44096452642, 2489)



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import itertools
from typing import Tuple
import numpy as np
import numpy.typing as npt

from pyscf.pbc import scf

from openfermion.resource_estimates.pbc.hamiltonian import (
    build_momentum_transfer_mapping,)


def get_df_factor(mat: npt.NDArray, thresh: float, verify_adjoint: bool = False
                 ) -> Tuple[npt.NDArray, npt.NDArray]:
    """Represent a matrix via non-zero eigenvalue vector pairs.

    Anything above thresh is considered non-zero

    Args:
        mat: Matrix to double factorize.
        thresh: Double factorization eigenvalue threshold
        verify_adjoint: Verify input matrix is Hermitian (Default value = False)

    Returns:
        Tuple eigen values and eigen vectors (lambda, V)

    """
    if verify_adjoint:
        assert np.allclose(mat, mat.conj().T)
    eigs, eigv = np.linalg.eigh(mat)
    normSC = np.sum(np.abs(eigs))
    ix = np.argsort(np.abs(eigs))[::-1]
    eigs = eigs[ix]
    eigv = eigv[:, ix]
    truncation = normSC * np.abs(eigs)
    to_zero = truncation < thresh
    eigs[to_zero] = 0.0
    eigv[:, to_zero] = 0.0
    idx_not_zero = np.where(~to_zero == True)[0]
    eigs = eigs[idx_not_zero]
    eigv = eigv[:, idx_not_zero]
    return eigs, eigv


class DFABKpointIntegrals:

    def __init__(self, cholesky_factor: npt.NDArray, kmf: scf.HF):
        """Class defining double factorized ERIs.

        We need to form the A and B objects which are indexed by Cholesky index
        n and momentum mode Q. This is accomplished by constructing rho[Q, n,
        kpt, nao, nao] by reshaping the cholesky object.  We don't form the
        matrix

        Args:
            cholesky_factor: Cholesky factor tensor that is
                [nkpts, nkpts, naux, nao, nao]
            kmf: pyscf k-object.  Currently only used to obtain the number of
                k-points.  must have an attribute kpts which len(self.kmf.kpts)
                returns number of kpts.
        """
        self.chol = cholesky_factor
        self.kmf = kmf
        self.nk = len(self.kmf.kpts)
        naux = 0
        for i, j in itertools.product(range(self.nk), repeat=2):
            naux = max(self.chol[i, j].shape[0], naux)
        self.naux = naux
        self.nao = cholesky_factor[0, 0].shape[-1]
        k_transfer_map = build_momentum_transfer_mapping(
            self.kmf.cell, self.kmf.kpts)
        self.k_transfer_map = k_transfer_map
        self.reverse_k_transfer_map = np.zeros_like(
            self.k_transfer_map)  # [kidx, kmq_idx] = qidx
        for kidx in range(self.nk):
            for qidx in range(self.nk):
                kmq_idx = self.k_transfer_map[qidx, kidx]
                self.reverse_k_transfer_map[kidx, kmq_idx] = qidx

        # set up for later when we construct DF
        self.df_factors = None
        self.a_mats = None
        self.b_mats = None

    def build_A_B_n_q_k_from_chol(self, qidx: int, kidx: int):
        """Builds matrices that are blocks in two momentum indices

              k  | k-Q |
            ------------
        k   |    |     |
        ----------------
        k-Q |    |     |
        ----------------

        where the off diagonal blocks are the ones that are populated.  All
        matrices for every Cholesky vector are constructed.

        Args:
            qidx: index for momentum mode Q.
            kidx: index for momentum mode K.

        Returns:
            Amat: The `A` matrix in DF (~ L + L^)
            Bmat: The `B` matrix in DF (~ i (L - L^)
        """
        k_minus_q_idx = self.k_transfer_map[qidx, kidx]
        naux = self.chol[kidx, k_minus_q_idx].shape[0]
        nmo = self.nao
        Amat = np.zeros((naux, 2 * nmo, 2 * nmo), dtype=np.complex128)
        Bmat = np.zeros((naux, 2 * nmo, 2 * nmo), dtype=np.complex128)
        if k_minus_q_idx == kidx:
            Amat[:, :nmo, :nmo] = self.chol[
                kidx, k_minus_q_idx]  # beacuse L_{pK, qK,n}= L_{qK,pK,n}^{*}
            Bmat[:, :nmo, :nmo] = 0.5j * (
                self.chol[kidx, k_minus_q_idx] -
                self.chol[kidx, k_minus_q_idx].conj().transpose(0, 2, 1))
        else:
            Amat[:, :nmo, nmo:] = (0.5 * self.chol[kidx, k_minus_q_idx]
                                  )  # [naux, nmo, nmo]
            Amat[:, nmo:, :nmo] = 0.5 * self.chol[kidx, k_minus_q_idx].conj(
            ).transpose(0, 2, 1)

            Bmat[:, :nmo, nmo:] = (0.5j * self.chol[kidx, k_minus_q_idx]
                                  )  # [naux, nmo, nmo]
            Bmat[:, nmo:, :nmo] = -0.5j * self.chol[kidx, k_minus_q_idx].conj(
            ).transpose(0, 2, 1)

        return Amat, Bmat

    def build_chol_part_from_A_B(
            self,
            kidx: int,
            qidx: int,
            Amats: npt.NDArray,
            Bmats: npt.NDArray,
    ) -> npt.NDArray:
        r"""Construct rho_{n, k, Q}.

        Args:
            kidx: k-momentum index
            qidx: Q-momentum index
            Amats: naux, 2 * nmo, 2 * nmo]
            Bmats: naux, 2 * nmo, 2 * nmo]

        Returns:
            cholesky factors: 3-tensors (k, k-Q)[naux, nao, nao],
                (kp, kp-Q)[naux, nao, nao]

        """
        k_minus_q_idx = self.k_transfer_map[qidx, kidx]
        nmo = self.nao
        if k_minus_q_idx == kidx:
            return Amats[:, :nmo, :nmo]
        else:
            return Amats[:, :nmo, nmo:] + -1j * Bmats[:, :nmo, nmo:]

    def double_factorize(self, thresh=None) -> None:
        """Construct a double factorization of the Hamiltonian.

        Iterate through qidx, kidx and get factorized Amat and Bmat for each
        Cholesky rank

        Args:
            thresh: Double factorization eigenvalue threshold
                (Default value = None)
        """
        if thresh is None:
            thresh = 1.0e-13
        if self.df_factors is not None:
            return self.df_factors

        nkpts = self.nk
        nmo = self.nao
        naux = self.naux
        self.amat_n_mats = np.zeros((nkpts, nkpts, naux, 2 * nmo, 2 * nmo),
                                    dtype=np.complex128)
        self.bmat_n_mats = np.zeros((nkpts, nkpts, naux, 2 * nmo, 2 * nmo),
                                    dtype=np.complex128)
        self.amat_lambda_vecs = np.empty((nkpts, nkpts, naux), dtype=object)
        self.bmat_lambda_vecs = np.empty((nkpts, nkpts, naux), dtype=object)
        for qidx, kidx in itertools.product(range(nkpts), repeat=2):
            Amats, Bmats = self.build_A_B_n_q_k_from_chol(qidx, kidx)
            naux_qk = Amats.shape[0]
            assert naux_qk <= naux
            for nc in range(naux_qk):
                amat_n_eigs, amat_n_eigv = get_df_factor(Amats[nc], thresh)
                self.amat_n_mats[kidx, qidx][nc, :, :] = (
                    amat_n_eigv @ np.diag(amat_n_eigs) @ amat_n_eigv.conj().T)
                self.amat_lambda_vecs[kidx, qidx, nc] = amat_n_eigs

                bmat_n_eigs, bmat_n_eigv = get_df_factor(Bmats[nc], thresh)
                self.bmat_n_mats[kidx, qidx][nc, :, :] = (
                    bmat_n_eigv @ np.diag(bmat_n_eigs) @ bmat_n_eigv.conj().T)
                self.bmat_lambda_vecs[kidx, qidx, nc] = bmat_n_eigs

    def get_eri(self, ikpts: list) -> npt.NDArray:
        """Construct (pkp qkq| rkr sks) via A and B tensors.

        Args:
            ikpts: list of four integers representing the index of the kpoint in
                self.kmf.kpts

        Returns:
            eris: ([pkp][qkq]|[rkr][sks])
        """
        ikp, ikq, ikr, iks = ikpts  # (k, k-q, k'-q, k')
        qidx = self.reverse_k_transfer_map[ikp, ikq]
        test_qidx = self.reverse_k_transfer_map[iks, ikr]
        assert test_qidx == qidx

        # build Cholesky vector from truncated A and B
        chol_val_k_kmq = self.build_chol_part_from_A_B(
            ikp, qidx, self.amat_n_mats[ikp, qidx], self.bmat_n_mats[ikp, qidx])
        chol_val_kp_kpmq = self.build_chol_part_from_A_B(
            iks, qidx, self.amat_n_mats[iks, qidx], self.bmat_n_mats[iks, qidx])

        return np.einsum(
            "npq,nsr->pqrs",
            chol_val_k_kmq,
            chol_val_kp_kpmq.conj(),
            optimize=True,
        )

    def get_eri_exact(self, ikpts: list) -> npt.NDArray:
        """Construct (pkp qkq| rkr sks) exactly from Cholesky vector.

        This is for constructing the J and K like terms needed for the one-body
        component lambda

        Args:
            ikpts: list of four integers representing the index of the kpoint in
                self.kmf.kpts

        Returns:
            eris: ([pkp][qkq]|[rkr][sks])
        """
        ikp, ikq, ikr, iks = ikpts
        return np.einsum(
            "npq,nsr->pqrs",
            self.chol[ikp, ikq],
            self.chol[iks, ikr].conj(),
            optimize=True,
        )



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt

from openfermion.resource_estimates.pbc.df.df_integrals import (
    DFABKpointIntegrals,)
from openfermion.resource_estimates.pbc.hamiltonian import (
    HamiltonianProperties,)


@dataclass
class DFHamiltonianProperties(HamiltonianProperties):
    """Store for return values of compute_lambda function

    Extension of HamiltonianProperties dataclass to also hold the number of
    retained eigenvalues (num_eig).
    """

    num_eig: int


def compute_lambda(hcore: npt.NDArray,
                   df_obj: DFABKpointIntegrals) -> DFHamiltonianProperties:
    """Compute lambda for double-factorized Hamiltonian.

    one-body term h_pq(k) = hcore_{pq}(k)
                            - 0.5 * sum_{Q}sum_{r}(pkrQ|rQqk)
                            + 0.5 sum_{Q}sum_{r}(pkqk|rQrQ)
    The first term is the kinetic energy + pseudopotential (or electron-nuclear)
    second term is from rearranging two-body operator into chemist charge-charge
    type notation, and the third is from the one body term obtained when
    squaring the two-body A and B operators.

    Arguments:
        hcore: List len(kpts) long of nmo x nmo complex hermitian arrays
        df_obj: DFABKpointIntegrals integral helper.

    Returns:
        ham_props: A HamiltonianProperties instance containing Lambda values for
            DF hamiltonian.
    """
    kpts = df_obj.kmf.kpts
    nkpts = len(kpts)
    one_body_mat = np.empty((len(kpts)), dtype=object)
    lambda_one_body = 0.0

    for kidx in range(len(kpts)):
        # matrices for - 0.5 * sum_{Q}sum_{r}(pkrQ|rQqk)
        # and  + 0.5 sum_{Q}sum_{r}(pkqk|rQrQ)
        h1_pos = np.zeros_like(hcore[kidx])
        h1_neg = np.zeros_like(hcore[kidx])
        for qidx in range(len(kpts)):
            # - 0.5 * sum_{Q}sum_{r}(pkrQ|rQqk)
            eri_kqqk_pqrs = df_obj.get_eri_exact([kidx, qidx, qidx, kidx])
            h1_neg -= (np.einsum("prrq->pq", eri_kqqk_pqrs, optimize=True) /
                       nkpts)
            # + 0.5 sum_{Q}sum_{r}(pkqk|rQrQ)
            eri_kkqq_pqrs = df_obj.get_eri_exact([kidx, kidx, qidx, qidx])
            h1_pos += np.einsum("pqrr->pq", eri_kkqq_pqrs) / nkpts

        one_body_mat[kidx] = hcore[kidx] + 0.5 * h1_neg + h1_pos
        one_eigs, _ = np.linalg.eigh(one_body_mat[kidx])
        lambda_one_body += np.sum(np.abs(one_eigs))

    lambda_two_body = 0.0
    num_eigs = 0
    for qidx in range(len(kpts)):
        for nn in range(df_obj.naux):
            first_number_to_square = 0
            second_number_to_square = 0
            # sum up p,k eigenvalues
            for kidx in range(len(kpts)):
                # A and B are W
                if df_obj.amat_lambda_vecs[kidx, qidx, nn] is None:
                    continue
                eigs_a_fixed_n_q = df_obj.amat_lambda_vecs[kidx, qidx,
                                                           nn] / np.sqrt(nkpts)
                eigs_b_fixed_n_q = df_obj.bmat_lambda_vecs[kidx, qidx,
                                                           nn] / np.sqrt(nkpts)
                first_number_to_square += np.sum(np.abs(eigs_a_fixed_n_q))
                num_eigs += len(eigs_a_fixed_n_q)
                if eigs_b_fixed_n_q is not None:
                    second_number_to_square += np.sum(np.abs(eigs_b_fixed_n_q))
                    num_eigs += len(eigs_b_fixed_n_q)

            lambda_two_body += first_number_to_square**2
            lambda_two_body += second_number_to_square**2

    lambda_two_body *= 0.25

    lambda_tot = lambda_one_body + lambda_two_body
    df_data = DFHamiltonianProperties(
        lambda_total=lambda_tot,
        lambda_one_body=lambda_one_body,
        lambda_two_body=lambda_two_body,
        num_eig=num_eigs,
    )
    return df_data



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from openfermion.resource_estimates.pbc.df.compute_df_resources import (
        _compute_cost,
        compute_cost,
    )


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_costing():
    nRe = 108
    lamRe = 294.8
    dE = 0.001
    LRe = 360
    LxiRe = 13031
    chi = 10
    betaRe = 16

    # (*The Li et al orbitals.*)
    nLi = 152
    lamLi = 1171.2
    LLi = 394
    LxiLi = 20115
    betaLi = 20

    res = _compute_cost(nRe, lamRe, dE, LRe, LxiRe, chi, betaRe, 2, 2, 2,
                        20_000)
    res = _compute_cost(nRe, lamRe, dE, LRe, LxiRe, chi, betaRe, 2, 2, 2,
                        res[0])
    # 48250, 22343175750, 8174
    assert np.isclose(res[0], 48250)
    assert np.isclose(res[1], 22343175750)
    assert np.isclose(res[2], 8174)

    res = _compute_cost(nRe, lamRe, dE, LRe, LxiRe, chi, betaRe, 3, 5, 1,
                        20_000)
    res = _compute_cost(nRe, lamRe, dE, LRe, LxiRe, chi, betaRe, 3, 5, 1,
                        res[0])
    # 53146, 24610371366, 8945
    assert np.isclose(res[0], 53146)
    assert np.isclose(res[1], 24610371366)
    assert np.isclose(res[2], 8945)

    res = _compute_cost(nLi, lamLi, dE, LLi, LxiLi, chi, betaLi, 2, 2, 2,
                        20_000)
    res = _compute_cost(nLi, lamLi, dE, LLi, LxiLi, chi, betaLi, 2, 2, 2,
                        res[0])
    # print(res) # 79212, 145727663004, 13873
    assert np.isclose(res[0], 79212)
    assert np.isclose(res[1], 145727663004)
    assert np.isclose(res[2], 13873)
    res = _compute_cost(nLi, lamLi, dE, LLi, LxiLi, chi, betaLi, 3, 5, 1,
                        res[0])
    # print(res) # 86042, 158292930114, 14952
    assert np.isclose(res[0], 86042)
    assert np.isclose(res[1], 158292930114)
    assert np.isclose(res[2], 14952)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_costing_helper():
    nRe = 108
    lamRe = 294.8
    dE = 0.001
    LRe = 360
    LxiRe = 13031
    chi = 10
    betaRe = 16

    # (*The Li et al orbitals.*)
    nLi = 152
    lamLi = 1171.2
    LLi = 394
    LxiLi = 20115
    betaLi = 20

    res = compute_cost(
        num_spin_orbs=nRe,
        lambda_tot=lamRe,
        num_aux=LRe,
        num_eig=LxiRe,
        kmesh=[2, 2, 2],
        dE_for_qpe=dE,
        chi=chi,
        beta=betaRe,
    )
    # 48250, 22343175750, 8174
    assert np.isclose(res.toffolis_per_step, 48250)
    assert np.isclose(res.total_toffolis, 22343175750)
    assert np.isclose(res.logical_qubits, 8174)

    res = compute_cost(
        num_spin_orbs=nRe,
        lambda_tot=lamRe,
        num_aux=LRe,
        num_eig=LxiRe,
        kmesh=[3, 5, 1],
        dE_for_qpe=dE,
        chi=chi,
        beta=betaRe,
    )
    # 53146, 24610371366, 8945
    assert np.isclose(res.toffolis_per_step, 53146)
    assert np.isclose(res.total_toffolis, 24610371366)
    assert np.isclose(res.logical_qubits, 8945)

    res = compute_cost(
        num_spin_orbs=nLi,
        lambda_tot=lamLi,
        num_aux=LLi,
        num_eig=LxiLi,
        kmesh=[3, 5, 1],
        dE_for_qpe=dE,
        chi=chi,
        beta=betaLi,
    )
    # print(res) # 79212, 145727663004, 13873
    assert np.isclose(res.toffolis_per_step, 86042)
    assert np.isclose(res.total_toffolis, 158292930114)
    assert np.isclose(res.logical_qubits, 14952)


if __name__ == "__main__":
    test_costing()



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from openfermion.resource_estimates.pbc.thc.compute_thc_resources import (
        _compute_cost,
        compute_cost,
    )


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_thc_resources():
    lam = 307.68
    dE = 0.001
    n = 108
    chi = 10
    beta = 16
    M = 350

    res = _compute_cost(n, lam, dE, chi, beta, M, 1, 1, 1, 20_000)
    # print(res) # 26205, 12664955115, 2069
    print(res)  # (80098, 38711603694, 17630)
    assert np.isclose(res[0], 80098)
    assert np.isclose(res[1], 38711603694)
    assert np.isclose(res[2], 17630)

    res = _compute_cost(n, lam, dE, chi, beta, M, 3, 3, 3, 20_000)
    # print(res)  # {205788, 99457957764, 78813
    print(res)  # (270394, 130682231382, 78815)
    assert np.isclose(res[0], 270394)
    assert np.isclose(res[1], 130682231382)
    assert np.isclose(res[2], 78815)

    res = _compute_cost(n, lam, dE, chi, beta, M, 3, 5, 1, 20_000)
    # print(res)  # 151622, 73279367466, 39628
    print(res)  #  (202209, 97728216327, 77517)
    assert np.isclose(res[0], 202209)
    assert np.isclose(res[1], 97728216327)
    assert np.isclose(res[2], 77517)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_thc_resources_helper():
    lam = 307.68
    dE = 0.001
    n = 108
    chi = 10
    beta = 16
    M = 350

    res = compute_cost(
        num_spin_orbs=n,
        lambda_tot=lam,
        thc_dim=M,
        kmesh=[1, 1, 1],
        dE_for_qpe=dE,
        chi=chi,
        beta=beta,
    )
    assert np.isclose(res.toffolis_per_step, 80098)
    assert np.isclose(res.total_toffolis, 38711603694)
    assert np.isclose(res.logical_qubits, 17630)

    res = compute_cost(
        num_spin_orbs=n,
        lambda_tot=lam,
        thc_dim=M,
        kmesh=[3, 3, 3],
        dE_for_qpe=dE,
        chi=chi,
        beta=beta,
    )
    assert np.isclose(res.toffolis_per_step, 270394)
    assert np.isclose(res.total_toffolis, 130682231382)
    assert np.isclose(res.logical_qubits, 78815)

    res = compute_cost(
        num_spin_orbs=n,
        lambda_tot=lam,
        thc_dim=M,
        kmesh=[3, 5, 1],
        dE_for_qpe=dE,
        chi=chi,
        beta=beta,
    )
    assert np.isclose(res.toffolis_per_step, 202209)
    assert np.isclose(res.total_toffolis, 97728216327)
    assert np.isclose(res.logical_qubits, 77517)



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from pyscf.pbc import gto, scf, cc

    from openfermion.resource_estimates.pbc.thc.factorizations.isdf import (
        solve_kmeans_kpisdf,)
    from openfermion.resource_estimates.pbc.thc.thc_integrals import (
        KPTHCDoubleTranslation,
        KPTHCSingleTranslation,
    )
    from openfermion.resource_estimates.pbc.hamiltonian.cc_extensions import (
        build_approximate_eris,)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
@pytest.mark.slow
def test_thc_helper():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.verbose = 0
    cell.mesh = [11] * 3
    cell.build()

    kmesh = [1, 1, 3]
    kpts = cell.make_kpts(kmesh)
    mf = scf.KRHF(cell, kpts)
    mf.kernel()

    exact_cc = cc.KRCCSD(mf)
    eris = exact_cc.ao2mo()
    exact_emp2, _, _ = exact_cc.init_amps(eris)

    approx_cc = cc.KRCCSD(mf)
    approx_cc.verbose = 0
    kpt_thc = solve_kmeans_kpisdf(
        mf,
        np.prod(cell.mesh),  # Use the whole grid to avoid any precision issues
        single_translation=False,
        use_density_guess=True,
        verbose=False,
    )

    helper = KPTHCDoubleTranslation(kpt_thc.chi, kpt_thc.zeta, mf)

    num_kpts = len(mf.kpts)
    for iq in range(num_kpts):
        for ik in range(num_kpts):
            ik_minus_q = helper.k_transfer_map[iq, ik]
            for ik_prime in range(num_kpts):
                ik_prime_minus_q = helper.k_transfer_map[iq, ik_prime]
                eri_thc = helper.get_eri(
                    [ik, ik_minus_q, ik_prime_minus_q, ik_prime])
                eri_exact = helper.get_eri_exact(
                    [ik, ik_minus_q, ik_prime_minus_q, ik_prime])
                assert np.allclose(eri_thc, eri_exact)

    eris_approx = build_approximate_eris(approx_cc, helper)
    emp2, _, _ = approx_cc.init_amps(eris_approx)
    assert np.isclose(emp2, exact_emp2)
    kpt_thc = solve_kmeans_kpisdf(mf,
                                  np.prod(cell.mesh),
                                  single_translation=True,
                                  verbose=False)
    helper = KPTHCSingleTranslation(kpt_thc.chi, kpt_thc.zeta, mf)
    for iq in range(num_kpts):
        for ik in range(num_kpts):
            ik_minus_q = helper.k_transfer_map[iq, ik]
            for ik_prime in range(num_kpts):
                ik_prime_minus_q = helper.k_transfer_map[iq, ik_prime]
                eri_thc = helper.get_eri(
                    [ik, ik_minus_q, ik_prime_minus_q, ik_prime])
                eri_exact = helper.get_eri_exact(
                    [ik, ik_minus_q, ik_prime_minus_q, ik_prime])
                assert np.allclose(eri_thc, eri_exact)

    eris_approx = build_approximate_eris(approx_cc, helper)
    emp2, _, _ = approx_cc.init_amps(eris_approx)
    assert np.isclose(emp2, exact_emp2)



#coverage:ignore
import os

import h5py
import numpy as np
import pytest

import openfermion.resource_estimates.integrals as int_folder
from openfermion.resource_estimates import (HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                                            thc)

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from openfermion.resource_estimates.molecule import load_casfile_to_pyscf


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_lambda():
    integral_path = int_folder.__file__.replace('__init__.py', '')
    thc_factor_file = os.path.join(integral_path, 'M_250_beta_16_eta_10.h5')
    eri_file = os.path.join(integral_path, 'eri_reiher.h5')
    with h5py.File(thc_factor_file, 'r') as fid:
        MPQ = fid['MPQ'][...]
        etaPp = fid['etaPp'][...]

    _, mf = load_casfile_to_pyscf(eri_file, num_alpha=27, num_beta=27)

    lambda_tot, nthc, _, _, _, _  = \
        thc.compute_lambda(mf, etaPp=etaPp, MPQ=MPQ)
    assert nthc == 250
    assert np.isclose(np.round(lambda_tot), 294)



#coverage:ignore
import dataclasses
import datetime
import math
from typing import Tuple, Iterator


@dataclasses.dataclass(frozen=True, unsafe_hash=True)
class MagicStateFactory:
    details: str
    physical_qubit_footprint: int
    rounds: int
    failure_rate: int


@dataclasses.dataclass(frozen=True, unsafe_hash=True)
class CostEstimate:
    physical_qubit_count: int
    duration: datetime.timedelta
    algorithm_failure_probability: float


@dataclasses.dataclass(frozen=True, unsafe_hash=True)
class AlgorithmParameters:
    physical_error_rate: float
    surface_code_cycle_time: datetime.timedelta
    logical_data_qubit_distance: int
    magic_state_factory: MagicStateFactory
    toffoli_count: int
    max_allocated_logical_qubits: int
    factory_count: int
    routing_overhead_proportion: float
    proportion_of_bounding_box: float = 1

    def estimate_cost(self) -> CostEstimate:
        """Determine algorithm single-shot layout and costs for given params.

        ASSUMES:
            - There is enough routing area to get needed data qubits to the
                factories as fast as the factories run.
            - The bottleneck time cost is waiting for magic states.
        """
        logical_storage = int(
            math.ceil(self.max_allocated_logical_qubits *
                      (1 + self.routing_overhead_proportion)))
        storage_area = logical_storage * _physical_qubits_per_logical_qubit(
            self.logical_data_qubit_distance)
        distillation_area = self.factory_count \
                          * self.magic_state_factory.physical_qubit_footprint
        rounds = int(self.toffoli_count / self.factory_count *
                     self.magic_state_factory.rounds)
        distillation_failure = self.magic_state_factory.failure_rate \
                             * self.toffoli_count
        data_failure = self.proportion_of_bounding_box \
                     * _topological_error_per_unit_cell(
            self.logical_data_qubit_distance,
            gate_err=self.physical_error_rate) * logical_storage * rounds

        return CostEstimate(physical_qubit_count=storage_area +
                            distillation_area,
                            duration=rounds * self.surface_code_cycle_time,
                            algorithm_failure_probability=min(
                                1., data_failure + distillation_failure))


def _topological_error_per_unit_cell(code_distance: int,
                                     gate_err: float) -> float:
    return 0.1 * (100 * gate_err)**((code_distance + 1) / 2)


def _total_topological_error(code_distance: int, gate_err: float,
                             unit_cells: int) -> float:
    return unit_cells * _topological_error_per_unit_cell(
        code_distance, gate_err)


def iter_known_factories(physical_error_rate: float
                        ) -> Iterator[MagicStateFactory]:
    if physical_error_rate == 0.001:
        yield _two_level_t_state_factory_1p1000(
            physical_error_rate=physical_error_rate)
    yield from iter_auto_ccz_factories(physical_error_rate)


def _two_level_t_state_factory_1p1000(physical_error_rate: float
                                     ) -> MagicStateFactory:
    assert physical_error_rate == 0.001
    return MagicStateFactory(
        details="https://arxiv.org/abs/1808.06709",
        failure_rate=4 * 9 * 10**-17,
        physical_qubit_footprint=(12 * 8) * (4) *
        _physical_qubits_per_logical_qubit(31),
        rounds=6 * 31,
    )


def _autoccz_factory_dimensions(l1_distance: int,
                                l2_distance: int) -> Tuple[int, int, float]:
    """Determine the width, height, depth of the magic state factory."""
    t1_height = 4 * l1_distance / l2_distance
    t1_width = 8 * l1_distance / l2_distance
    t1_depth = 5.75 * l1_distance / l2_distance

    ccz_depth = 5
    ccz_height = 6
    ccz_width = 3
    storage_width = 2 * l1_distance / l2_distance

    ccz_rate = 1 / ccz_depth
    t1_rate = 1 / t1_depth
    t1_factories = int(math.ceil((ccz_rate * 8) / t1_rate))
    t1_factory_column_height = t1_height * math.ceil(t1_factories / 2)

    width = int(math.ceil(t1_width * 2 + ccz_width + storage_width))
    height = int(math.ceil(max(ccz_height, t1_factory_column_height)))
    depth = max(ccz_depth, t1_depth)

    return width, height, depth


def iter_auto_ccz_factories(physical_error_rate: float
                           ) -> Iterator[MagicStateFactory]:
    for l1_distance in range(5, 25, 2):
        for l2_distance in range(l1_distance + 2, 41, 2):
            w, h, d = _autoccz_factory_dimensions(l1_distance=l1_distance,
                                                  l2_distance=l2_distance)
            f = _compute_autoccz_distillation_error(
                l1_distance=l1_distance,
                l2_distance=l2_distance,
                physical_error_rate=physical_error_rate)

            yield MagicStateFactory(
                details=
                f"AutoCCZ({physical_error_rate=},{l1_distance=},{l2_distance=}",
                physical_qubit_footprint=w * h *
                _physical_qubits_per_logical_qubit(l2_distance),
                rounds=d * l2_distance,
                failure_rate=f,
            )


def _compute_autoccz_distillation_error(l1_distance: int, l2_distance: int,
                                        physical_error_rate: float) -> float:
    # Level 0
    L0_distance = l1_distance // 2
    L0_distillation_error = physical_error_rate
    L0_topological_error = _total_topological_error(
        unit_cells=100,  # Estimated 100 for T injection.
        code_distance=L0_distance,
        gate_err=physical_error_rate)
    L0_total_T_error = L0_distillation_error + L0_topological_error

    # Level 1
    L1_topological_error = _total_topological_error(
        unit_cells=1100,  # Estimated 1000 for factory, 100 for T injection.
        code_distance=l1_distance,
        gate_err=physical_error_rate)
    L1_distillation_error = 35 * L0_total_T_error**3
    L1_total_T_error = L1_distillation_error + L1_topological_error

    # Level 2
    L2_topological_error = _total_topological_error(
        unit_cells=1000,  # Estimated 1000 for factory.
        code_distance=l2_distance,
        gate_err=physical_error_rate)
    L2_distillation_error = 28 * L1_total_T_error**2
    L2_total_CCZ_or_2T_error = L2_topological_error + L2_distillation_error

    return L2_total_CCZ_or_2T_error


def _physical_qubits_per_logical_qubit(code_distance: int) -> int:
    return (code_distance + 1)**2 * 2


def cost_estimator(num_logical_qubits,
                   num_toffoli,
                   physical_error_rate=1.0E-3,
                   portion_of_bounding_box=1.):
    """
    Produce best cost in terms of physical qubits and real run time based on
    number of toffoli, number of logical qubits, and physical error rate.

    """
    best_cost = None
    best_params = None
    for factory in iter_known_factories(
            physical_error_rate=physical_error_rate):
        for logical_data_qubit_distance in range(7, 35, 2):
            params = AlgorithmParameters(
                physical_error_rate=physical_error_rate,
                surface_code_cycle_time=datetime.timedelta(microseconds=1),
                logical_data_qubit_distance=logical_data_qubit_distance,
                magic_state_factory=factory,
                toffoli_count=num_toffoli,
                max_allocated_logical_qubits=num_logical_qubits,
                factory_count=4,
                routing_overhead_proportion=0.5,
                proportion_of_bounding_box=portion_of_bounding_box)
            cost = params.estimate_cost()
            if cost.algorithm_failure_probability > 0.1:
                continue
            if best_cost is None or cost.physical_qubit_count * cost.duration \
                < best_cost.physical_qubit_count * best_cost.duration:
                best_cost = cost
                best_params = params
    return best_cost, best_params
